# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

#  Screenlets Main Module - (c) 2008 by Natan Yellin (Aantn) <aantny@gmail.com>
#  and Whise (Helder Fraga) <helder.fraga@hotmail.com>
#
#  Original author RYX (aka Rico Pfaus) <ryx@ryxperience.com> 

##@mainpage
#
##@section intro_sec General Information
#
# INFO:
# - Universal-Applets are small owner-drawn applications that can be described as
#  "the virtual representation of things lying/standing around on your desk".
#   Sticknotes, clocks, rulers, etc. the possibilities are endless. The goal of 
#   the UA is to simplify the creation of fully themeable mini-apps that
#   each solve basic desktop-work-related needs and generally improve the 
#   usability and eye-candy of the modern Linux-desktop. UA are fork from Screenlets 
#   project.
#

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

#-------------------------------------------------------------------------------
# Imports!
#-------------------------------------------------------------------------------

import os
import sys
from optparse import OptionParser

import gtk
import gettext

#-------------------------------------------------------------------------------
# Find Install Prefix
# Note: It's important to do this before we import the Screenlets submodules
#-------------------------------------------------------------------------------
try:
    INSTALL_PREFIX = open('/etc/universal-applets/prefix').read().strip()
except IOError:
    INSTALL_PREFIX = '/usr'

#-------------------------------------------------------------------------------
# Parse command line options
# Note: This must be done before we import any submodules
# TODO: Fix up command line parsing. Right now, the same options are parsed for
#	both screenlets, the manager, the daemon, and melange.
#-------------------------------------------------------------------------------
# Define command line related constants.
# These are used as the default values for command line options that aren't overriden
LOG_DISABLED = False				# Disable log
LOG_LEVEL = 4						# TODO: change to 3 for stable version to show only warning, error and critical messages
LOG_OUTPUT = "FILE"					# default output, allowed FILE, STDERR, STDOUT
LOG_NAME = "universal-applets"		# Log name
LOG_FILE = "/tmp/%s.log" % os.path.basename(sys.argv[0]) # full path to log file; only used if LOG_OUTPUT is FILE

SESSION_NAME = "default"
SESSION_REPLACE = "False"
SERVER_NAME = "Melange"

# Create the parser
parser = OptionParser()
# Add options used by all of the various modules
parser.add_option("-l", "--logging-level", dest = "LOG_LEVEL", default = LOG_LEVEL,
	help = ("set logging level.\n0 - log disabled, 1 - only critical errors, 2 - all errors,\
	3 - all errors and warnings, 4 - all errors, warnings, and info messages, 5 - all messages \
	including debug output.\ndefault: %s") %(LOG_LEVEL))
parser.add_option("-f", "--logging-file", dest = "LOG_FILE", default = LOG_FILE,
	help = ("write log to LOG_FILE. Default: %s") %(LOG_FILE))
parser.add_option("-s", "--server", dest = "SERVER_NAME", default = SERVER_NAME,
	help = "server name. default options are Melange and Sidebar")
parser.add_option("-o", "--output", dest = "LOG_OUTPUT", default = LOG_OUTPUT,
	help = ("set output. allowed outputs: FILE, STDOUT and STDERR. Default: %s") %(LOG_OUTPUT))
parser.add_option("-q", "--quiet", action="store_true", dest = "LOG_DISABLED",
	default = LOG_DISABLED, help = "Disable log. The same as log level 0")
parser.add_option("-r", "--replace", action="store_true", dest = "SESSION_REPLACE",
	default = SESSION_REPLACE, help = "Replace old session. Default: %s" %(SESSION_REPLACE))
parser.add_option("-e", "--session", action="store_true", dest = "SESSION_NAME",
	default = SESSION_NAME, help = "Name of session. Default: %s" %(SESSION_NAME))
# Parse the options and store them in global module variables
COMMAND_LINE_OPTIONS, COMMAND_LINE_ARGS = parser.parse_args()

#-------------------------------------------------------------------------------
# Finish with the imports and import all of the submodules
#-------------------------------------------------------------------------------
from . import logger
from . import utils
from . import services
from . import plugin_loader as Plugins
from .applet import AppletView
from .screenlet import Screenlet
from .singleton import SingletonApplication
from .menu import DefaultMenuItem, add_menuitem, create_menu_from_file
from .widgets import ShapedWidget


#-------------------------------------------------------------------------------
# CONSTANTS
#-------------------------------------------------------------------------------

#-------------------------------------
# Meta info
#-------------------------------------

APP_NAME = "Universal-Applets"
COMMENTS = "Universal-Applets is a widget framework for Linux. Universal-Applets tries to include some compatibility with other widget frameworks, like web widgets and super karamba themes"
COPYRIGHT = "Natan Yellin (Aantn) <aantny@gmail.com>, see AUTHORS for authors of Screenlets, base for Universal-Applets"
VERSION = "0.1"
WEBSITE = "http://www.theesylum.com"
THIRD_PARTY_DOWNLOAD = "http://screenlets.org/index.php/Category:UserScreenlets" #TODO: Change ?

#-------------------------------------
# Credits
#-------------------------------------

AUTHORS = ["Natan Yellin (Aantn) <aantny@gmail.com>", "Check AUTHORS file for original authors"]
TRANSLATORS = "Check https://translations.launchpad.net/universal-applets/ for the details" #TODO: set translation
DOCUMENTERS = ["Documentation generated by doxygen"]

#-------------------------------------
# Directories
#-------------------------------------

# the temporary directory used for various file operations
TEMP_DIR = '/tmp/screenlets'

# file used to monitor running screenlets
TEMP_FILE	= TEMP_DIR + '/screenlets.' + os.environ['USER'] + '.running'

# the global directory where the screenlets are installed 
PATH = INSTALL_PREFIX + '/share/screenlets'

# the per-user directory where additional screenlets can be installed
USER_PATH = os.environ['HOME'] + '/.screenlets'

# A list containing all of the paths to search for individual screenlets
SCREENLETS_PATH = [USER_PATH, PATH]

# the config directory for screenlets
CONFIG_DIR = os.environ['HOME'] + '/.config/Screenlets'


#-------------------------------------
# Other
#-------------------------------------
DEBUG_MODE = True
log=logger.get_default_logger()

#-------------------------------------------------------------------------------
# Setup Translations
#-------------------------------------------------------------------------------
gettext.textdomain('universal-applets')
gettext.bindtextdomain('universal-applets', INSTALL_PREFIX +  '/share/locale')

_ = gettext.gettext

# ------------------------------------------------------------------------------
# Public Module Functions
# ------------------------------------------------------------------------------

def alpha_screen_changed (widget, screen=None):
	"""set colormap for window"""
	if screen is None:
		screen = widget.get_screen()
	map = screen.get_rgba_colormap()
	if map is None:
		map = screen.get_rgb_colormap()
	widget.set_colormap(map)

def create_empty_bitmap (width, height):
	data = ''.zfill(width*height)
	return gtk.gdk.bitmap_create_from_data(None, data, width, height)

# the new recommended way of launching a screenlet from the "outside"
def launch_screenlet (name, debug=False):
	"""Launch a screenlet, either through its service or by launching a new
	process of the given screenlet. Name has to be the name of the Screenlet's
	class without trailing 'Screenlet'.
	NOTE: we could only launch the file here"""
	# check for service
	if services.service_is_running(name):
		# add screenlet through service, if running
		srvc = services.get_service_by_name(name)
		if srvc:
			try:
				srvc.add('')	# empty string for auto-creating ID
				return True
			except Exception, ex:
				log.error("Error while adding instance by service: %s" % ex)
	# service not running or error? launch screenlet's file
	path = utils.find_first_screenlet_path(name)
	if path:
		# get full path of screenlet's file
		slfile = path + '/' + name + 'Screenlet.py'
		# launch screenlet as separate process
		log.info("Launching Screenlet from: %s" % slfile)
		if debug:
			print "Logging output goes to: $HOME/.config/Screenlets/%sScreenlet.log" % name
			out = '$HOME/.config/Screenlets/%sScreenlet.log' % name
		else:
			out = '/dev/null'
		os.system('python -u -t %s > %s &' % (slfile, out))
		return True
	else:
		log.warning("Screenlet '%s' could not be launched." % name)
		return False

def show_message (screenlet, message, title=''):
	"""Show a message for the given Screenlet (may contain Pango-Markup).
	If screenlet is None, this function can be used by other objects as well."""
	if screenlet is None:
		md = gtk.MessageDialog(None, type=gtk.MESSAGE_INFO, 
			buttons=gtk.BUTTONS_OK)
		md.set_title(title)
	else:
		md = gtk.MessageDialog(None, type=gtk.MESSAGE_INFO, 
			buttons=gtk.BUTTONS_OK)
		md.set_title(screenlet.__name__)
	md.set_markup(message)
	md.run()
	md.destroy()

def show_question (screenlet, message, title=''):
	"""Show a question for the given Screenlet (may contain Pango-Markup)."""
	if screenlet is None:
		md = gtk.MessageDialog(None, type=gtk.MESSAGE_QUESTION, 
			buttons=gtk.BUTTONS_YES_NO)
		md.set_title(title)
	else:
		md = gtk.MessageDialog(None, type=gtk.MESSAGE_QUESTION, 
			buttons=gtk.BUTTONS_YES_NO)
		md.set_title(screenlet.__name__)
	md.set_markup(message)
	response = md.run()
	md.destroy()
	if response == gtk.RESPONSE_YES:
		return True
	return False

def show_error (screenlet, message, title='Error'):
	"""Show an error for the given Screenlet (may contain Pango-Markup)."""
	if screenlet is None:
		md = gtk.MessageDialog(None, type=gtk.MESSAGE_ERROR, 
			buttons=gtk.BUTTONS_OK)
		md.set_title(title)
	else:
		md = gtk.MessageDialog(screenlet.window, type=gtk.MESSAGE_ERROR, 
			buttons=gtk.BUTTONS_OK)
		md.set_title(screenlet.__name__)
	md.set_markup(message)
	md.run()
	md.destroy()

def fatal_error (message):
	"""Raise a fatal error to stdout and stderr and exit with an errorcode."""
	msg = 'FATAL ERROR: %s\n' % message
	sys.stdout.write(msg)
	sys.stderr.write(msg)
	sys.exit(1)

def non_fatal_deprecated_warning ():
	"""Warn the user that the screenlet they are using is deprecated."""
	sys.stdout.write("WARNING: The screenlet you are using needs to be updated\nPlease inform the developer")

# LEGACY support: functions that are not used any longer (raise fatal error)

def create_new_instance (name):
	fatal_error("This screenlet seems to be written for an older version of the framework. Please download a newer version of the %s." % name)

