# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Applet and AppletView class.
# (c) 2008 Natan Yellin (Aantn) <aantny@gmail.com>

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

import sys

import cairo
import dbus
import gettext
import pygtk
pygtk.require('2.0')
import gtk
import gobject
import wnck

import screenlets
from .theme import Theme
from . import logger

# set up translations
gettext.textdomain('screenlets')
gettext.bindtextdomain('screenlets', screenlets.INSTALL_PREFIX +  '/share/locale')

_ = gettext.gettext

# Global setting for whether the default screen is composited
def composited_changed (screen, event):
	SCREEN_IS_COMPOSITED = screen.is_composited()
	
screen = gtk.gdk.screen_get_default()
screen.connect("composited-changed", composited_changed, screen)
SCREEN_IS_COMPOSITED = screen.is_composited()

# set up logger
log = logger.get_default_logger()
		
class AppletView (gtk.VBox):
	"""Class used to "wrap" a gtk.Widget as an applet."""
	#TODO: Subclass gtk.Bin
	
	def __init__ (self, id, width=-1, height=-1, size_type="scalable",
		server_name='Melange'):
		"""
		Constructor.
		id: A unique string used to identify the applet.
		width: The applet's width. This is interpreted according to size_type.
		height: The applet's height. This is interpreted according to size_type.
		size_type: Determines the meaning of width and height.
			May be one of the following values:
			"scalable"- The applet may be resized as long as the width/height ratio is kept.
			"static"- The applet should be displayed at exactly the size requested.
			"static-width"-	The applet's width should remain static, and the server may change the height.
			"static-height"- The applet's height should remain static, and the server may change the width.
			"dynamic"- The applet may be resized to any size.
		server: The server to display the applet in.
		"""
		super(AppletView, self).__init__()
		
		# Store parameters as attributes
		self.id			= id			# a unique string id used to communicate with the server
		self.width		= width			# a width that we'll suggest to the server
		self.height		= height		# a height that we'll suggest to the server
		self.size_type	= size_type		# the size type. see documentation in src/bin/melange
		self.server_name = server_name	# the server that we'll connect to
		
		# Initialize other attributes
		self.flags	 = {}			# a list of flags obtained from the server
		self.plug	 = None			# a GtkPlug or None when the applet is deactivated
		self.dragged = False		# True when the mouse is down over the applet
		self.running_servers = None	# a list of all running servers 
		
		# Setup DBUS
		self.bus = dbus.SessionBus()
		self.iface	= None	# a DBUS interface for the currently used server
	
	# ----------------------------------------------------
	# Public Methods
	# ----------------------------------------------------
	
	def get_active (self):
		"""Returns True if the applet is currently in plug mode"""
		if self.plug is not None:
			return True
		else:
			return False
	
	def get_running_servers (self):
		"""This function returns list of all running servers"""
		self.running_servers = []
		for service in self.bus.list_names():
			if service[0:21] == "org.UniversalApplets.":
				# for now, manually filter out the manager and dameon
				# we'll fix this later on in a more elegant manner 
				if service != "org.UniversalApplets.UniversalAppletsManager" \
				and service != "org.UniversalApplets.Daemon":
					self.running_servers.append(service[21:])
		return self.running_servers

	def get_flag (self, flag, default=False):
		"""Check a flag from the server.
		If the flag doesn't exist, default is returned"""
		if flag in self.flags:
			return self.flags[flag]
		else:
			return default
	
	def get_server (self):
		"""Returns the name of the currently set server.
		Note that the server isn't guaranteed to be running."""
		return self.server_name
	
	def get_size (self):
		"""This function returns a tuple with width, height, and size type. See constructor ( __init__ ) for possible types of size"""
		return (self.width, self.height, self.size_type)
		
	def get_widget (self):
		"""Returns the child widget or None."""
		if self.plug is not None:
			widget = self.plug
		else:
			widget = self
			
		if len(widget.get_children()) > 0:
			return widget.get_children()[0]
		else:
			return None
	
	def set_active (self, active=True):
		"""Activate/Deactivate the plug mode."""
		if active:
			log.debug("Activating plug.")
			# connect to the server via dbus
			self.__connect_to_server()
			
			# create the gtk.Plug
			self.plug = gtk.Plug(long(0))
			self.plug.connect("button-press-event", self.plug_button_press_event)
			self.plug.connect("focus-in-event", self.plug_focus_in_event)
			self.plug.connect("enter-notify-event", self.plug_enter_notify_event)
			self.plug.connect("leave-notify-event", self.plug_leave_notify_event)
			screenlets.alpha_screen_changed(self.plug)
			
			# hide the normal widget and reparent the applet
			self.hide()
			if len(self.get_children()) > 0:
				self.get_children()[0].reparent(self.plug)
			self.plug.show()
			
			# register with the server
			self.iface.add_applet(self.id, self.plug.get_id(), self.width, self.height, "scalable")
			
			# get a list of flags from the server
			self.flags = {}
			for key, value in self.iface.get_all_server_flags(self.id).items():
				self.flags[key] = value
				
		elif self.plug is not None and self.iface is not None:
			log.debug("Deactivating plug")
			#self.iface.remove_view(self.id, view.id)
			if len(self.plug.get_children()) > 0:
				self.plug.get_children()[0].reparent(self)
			
			self.plug.destroy()
			self.plug = None
			self.iface = None
			self.show()
	
	def set_server (self, server_name):
		"""Sets the applet's server."""
		# disconnect from the current server
		if self.iface is not None:
			self.set_active(False)
			reactivate = True
		else:
			reactivate = False
		
		self.server_name = server_name
			
		if reactivate:
			self.set_active(True)
	
	def set_size (self, width=-1, height=-1, size_type=""):
		"""
		Sets an applet's size and/or size type.
		
		width: A recommended width or -1 to preserve the current width. This will be
			interpreted according to size_type.
		height A recommended height or -1 to preserve the current height. This will be
			interpreted according to size_type.
		size_type: Determines the meaning of width and height. The below values are all
			acceptable, or "" may be given to preserve the current size type:
			"scalable"- The applet may be resized as long as the width/height ratio is kept.
			"static"- The applet should be displayed at exactly the size requested.
			"static-width"-	The applet's width should remain static, and the server may change the height.
			"static-height"- The applet's height should remain static, and the server may change the width.
			"dynamic"- The applet may be resized to any size.
		"""
		self.width = width
		self.height = height
		self.size_type = size_type
		
		if self.iface is not None:
			self.iface.set_applet_size(self.id, self.width, self.height, "")
		
	def set_shape (self, bitmap, input_only=True):
		"""Sets the applet's shape."""
		if self.get_flag("supports-shaping"):
			if self.plug is None:
				dest = self.window
			else:
				dest = self.plug.window
			
			if dest is None:
				log.debug("Could not set window shape.")
			else:
				if input_only:
					log.debug("Setting input shape.")
					dest.input_shape_combine_mask(bitmap, 0, 0)
				else:
					log.debug("Setting window shape.")
					dest.shape_combine_mask(bitmap, 0, 0)
				
			if self.plug is not None:
				self.iface.send_message(self.id, "shape-changed", 0)
		
	# -------------------------------------------------
	# Gtk Event Handlers
	# -------------------------------------------------
	
	def plug_button_press_event (self, widget, event):
		"""Event handler for button presses.
		It's important to ungrab the pointer and then to let the
		server know that the mouse was clicked."""
		if self.get_flag("send-button-press-event"):
			gtk.gdk.pointer_ungrab(event.time)
			self.iface.send_message(self.id, "button-press-event", 0)
	
	def plug_focus_in_event (self, widget, event):
		"""Event handler for button releases."""
		if self.get_flag("send-focus-in-event"):
			self.iface.send_message(self.id, "focus-in-event", 0)
	
	def plug_enter_notify_event (self, widget, event):
		"""Event handler for when the mouse enters the window."""
		if self.get_flag("send-enter-notify-event"):
			self.iface.send_message(self.id, "enter-notify-event", 0)
	
	def plug_leave_notify_event (self, widget, event):
		"""Event handler for when the mouse leaves the window."""
		if self.get_flag("send-leave-notify-event"):
			self.iface.send_message(self.id, "leave-notify-event", 0)
	
	def socket_button_release_event (self, id, x, y, state,
		button, x_root, y_root):
		"""Handler for the button_release_event signal emitted by
		the server when it wishes to forward a button-release-event
		to the applet.""" 
		# check if the signal that the server emitted is meant for this applet
		# TODO: Do this automatically using DBUS.Interface.connect_to_signal's
		# arg0 parameter. (Which seems to be broken.)
		if id == self.id:
			child = self.get_widget()
			if child is not None and child.window is not None:
				event = gtk.gdk.Event(gtk.gdk.BUTTON_RELEASE)
				event.window = child.window
				event.x = x
				event.y = y
				event.state = state
				event.button = button
				event.x_root = x_root
				event.y_root = y_root
				event.put()
		
	# -------------------------------------------------
	# Private Methods
	# -------------------------------------------------
	
	def __connect_to_server (self):
		"""Connect to an applet server."""
		bus_name	= "org.UniversalApplets." + self.server_name
		path		= "/org/UniversalApplets/" + self.server_name
		iface		= bus_name
		
		proxy_obj = self.bus.get_object(bus_name, path)
			
		if proxy_obj is not None:
			self.iface = dbus.Interface(proxy_obj, iface)
			self.iface.connect_to_signal("button_release_event",
				self.socket_button_release_event)
