# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Screenlets Backend Module
# (c) 2007 Rico Pfaus (RYX) <ryx@ryxperience.com>

# INFO:
# - The backend offers an abstracted way of saving a Screenlet's data
#
# TODO: 
# - add "type"-argument to save_option and read_option to be able to correctly
#   set the values in GconfBackend (instead of storing only strings).
#

import glob
import os
import gobject
import gettext
from xdg import BaseDirectory
try:
	import gconf
except:
	print _("GConf python module not found. GConf settings backend is disabled.")

gettext.textdomain('screenlets')
gettext.bindtextdomain('screenlets', '/usr/share/locale')

_ = gettext.gettext


class ScreenletsBackend:
	"""The backend performs the loading/saving of the 'key=value'-strings. 
	Extend this superclass to implement different saving-backends."""
	
	def __init__ (self):
		pass
	
	def delete_instance (self, id):
		"""Delete an instance's configuration by its id."""
		pass
	
	def flush (self):
		"""Immediately store all values to disk (in case the backend doesn't
		save in realtime anyway."""
		pass
	
	def load_option (self, id, name):
		"""Load one option for the instance with the given id."""
		pass
	
	def load_instance (self, id):
		"""Load all options for the instance with the given id."""
		pass
	
	def save_option (self, id, name, value):
		"""Save one option for the instance with the given id."""
		pass
	

class GconfBackend (ScreenletsBackend):
	"""Backend for storing settings in the GConf registry"""
	
	gconf_dir = '/apps/screenlets/'
	
	def __init__ (self):
		ScreenletsBackend.__init__(self)
		print _('GConfBackend: initializing')
		self.client = gconf.client_get_default()
	
	def delete_instance (self, id):
		"""Delete an instance's configuration by its id."""
		os.system('gconftool-2 --recursive-unset ' + self.key + id)
		return True
	
	def flush (self):
		"""Immediately store all values to disk (in case the backend doesn't
		save in realtime anyway."""
		pass	#No need, GConf saves in realtime

	def load_option (self, id, name):
		"""Load one option for the instance with the given id."""
		return self.client.get_string(self.gconf_dir + id + '/' + name)
	
	def load_instance (self, id):
		"""Load all options for the instance with the given id."""
		keys = []
		vals = []
		for i in self.client.all_entries(self.gconf_dir + id):
			keys.append(i.key.split('/')[4])
			vals.append(self.client.get_string(i.key))
		return dict(zip(keys, vals))
		return None
	
	def save_option (self, id, name, value):
		"""Save one option for the instance with the given id."""
		self.client.set_string(self.gconf_dir + id + '/' + name, value)
		print _('Saved option %s%s/%s = %s') % (self.gconf_dir, id, name, value)


class CachingBackend (ScreenletsBackend):
	"""A backend that stores the settings in arrays and saves after a short 
	interval to avoid overhead when multiple values are set within a short time. 
	The data gets saved into $HOME/.config/Screenlets/<Screenletname>/, in a 
	file for each element (named like its id with the extension '.ini')."""
	
	# Class attributes
	__delay_time = 3000		# delay to wait before performing save
	
	def __init__ (self, path):
		"""Constructor"""
		ScreenletsBackend.__init__(self)
		
		# find the path containing saved instances
		self.path = BaseDirectory.load_first_config(path)
		if self.path is None:
			self.path = BaseDirectory.save_config_path(path)
			
		self.__instances	= {}	# dict of ids and loaded instances
		
		# timeout and related attributes
		self.__timeout 		= None	# the id of the timeout-function
		self.__queue 		= []	# list with ids of instances that need saving
		
		# load a dict of ids:saved files in self.path
		self.__configs = self.__load_configs()
	
	def delete_instance (self, id):
		"""Delete an instance from the list and from the filesystem."""
		# remove the instance from the cache and the queue
		if self.__instances.has_key(id):
			self.__instances.pop(id)
		while self.__queue.count(id) > 0:
			self.__queue.remove(id)
		
		# delete the config file
		config_file = self.__configs.pop(id)
		try:
			os.remove(config_file)
			return True
		except Exception,ex:
			print ex
			return False
	
	def flush (self):
		"""Immediately save all pending data."""
		self.__save_cache()
	
	def get_configs (self):
		"""Return a the ids of all saved instances"""
		return self.__configs.keys()
		
	def load_instance (self, id):
		"""Load and return all options for the instance with the given id."""
		# if the instance wasn't already loaded then load it
		if self.__instances.has_key(id) == False:
			self.__instances[id] = {}
			if self.__configs.has_key(id):
				file = open(self.__configs[id], 'r')
				lines = file.readlines()
				# read all options for this element from file
				for line in lines:
					parts = line[:-1].split('=', 1)
					if len(parts) > 1:
						self.__instances[id][parts[0]] = parts[1]
				file.close()
		# return the dict of options
		return self.__instances[id]
	
	def load_option (self, id, name, default=None):
		"""Load and return option from the backend (returned as str).
		If the option doesn't exist, default is returned."""
		options = self.load_instance(id)
		if options.has_key(name):
			return options[name]
		else:
			return default
	
	def save_option (self, id, name, value):
		"""Save option for an instance to cache and start saving-timeout 
		for that element (value must be of type string)."""
		# add the instance to the queue and list of instances
		if self.__queue.count(id) == 0:
			self.__queue.append(id)			
		if self.__instances.has_key(id) == False:
			self.__instances[id] = {}
		if self.__configs.has_key(id) == False:
			self.__configs[id] = self.path + id + '.ini'
		# set the option in cache
		self.__instances[id][name] = str(value)
		
		# reset the timeout
		if self.__timeout:
			gobject.source_remove(self.__timeout)
		self.__timeout = gobject.timeout_add(self.__delay_time, 
			self.__save_cache)
	
	# Private methods
	
	def __load_configs (self):
		"""Load a list of all cached files from path."""
		configs = {}
		plen = len(self.path)
		files = glob.glob(self.path + '*.ini')
		for file in files:
			id = file[plen:-4]
			configs[id] = file
		return configs
	
	def __save_cache (self):
		"""Save the cache (for all pending instances in queue) to self.path."""
		# loop through all instances in queue:
		for id in self.__queue:
			# if the element in the queue does not exists, ignore it
			if self.__instances.has_key(id) and len(self.__instances[id]) > 0:
				file = open(self.__configs[id], 'w')
				for name, value in self.__instances[id].items():
					file.write("%s=%s\n" % (name, value))
				file.close()
		# clear queue
		self.__queue = []
		# IMPORTANT: return False to stop the timeout-function
		return False
