# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Universal Applets Melange Module
# (c) 2008 Natan Yellin (Aantn) <aantny@gmail.com>

# TODO: More cleanups.
# Please see the TODO file for long term goals.

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

import os
import gettext

import pygtk
pygtk.require('2.0')
import gtk
import cairo
import gobject

import screenlets
import screenlets.utils
import screenlets.sensors
import screenlets.logger
from screenlets.drawing import Drawing
from screenlets.menu import DefaultMenuItem, add_menuitem
from screenlets.options import *


log = screenlets.logger.get_default_logger()


_ = gettext.gettext


class ToplevelContainer (EditableOptions, Drawing):
	"""A toplevel window which displays an applet using a gtk.Plug."""
	
	# constants
	FADE_TIME		= 60	# milliseconds that the timeout to fade the buttons is called
	BUTTONS_WIDTH	= 16	# the width of both the border and the floating buttons
	
	BOOST = 1.5
	
	LOW_GRAVITY = 1.0
	MEDIUM_GRAVITY = 1.5
	HIGH_GRAVITY = 2.0
	
	LOW_BOUNCE = -0.2
	MEDIUM_BOUNCE = -0.7
	HIGH_BOUNCE = -0.9
	
	LOW_FRICTION = 0.999
	MEDIUM_FRICTION = 0.96
	HIGH_FRICTION = 0.9
	
	def __init__ (self, plug_id, applet_id, server, width=100, height=100,
		size_type= "scalable", backend=None):
		"""Constructor"""
		super(ToplevelContainer, self).__init__(backend, applet_id, True)
		
		# init attributes from parameter
		self.id 		= applet_id # a unique string used to identify the applet
		self.width 		= width		# width
		self.height 	= height	# height
		self.size_type	= size_type	# size type (see src/bin/melange) which is currently ignored
		self.server 	= server	# the Melange object
		self.backend 	= backend	# melange's backend
		
		# load default values from file
		ini = utils.IniReader()
		ini.load_conf (os.environ['HOME'] + '/.screenlets' + '/config.ini')
		
		# TODO: Add code to options.py to support default option settings
		def parse_bool (name, default=False):
			if ini and ini.has_section('Options'):
				if ini.get_option(name, section='Options') == "True":
					return True
				elif ini.get_option(name, section='Options') == "False":
					return False
			return default
		
		# TODO: Add support for default options directly into the option module so that this isn't necessary
		lock_position_default	= parse_bool('Lock') 					# lock the container's position on screen. attempts to move it are ignored.
		is_sticky_default		= parse_bool('Sticky', True)			# show the container on all desktops
		is_widget_default		= parse_bool('Widget', False)			# show the container on a special widget layer
		keep_above_default		= parse_bool('Keep_above', True)		# keep the container above other windows
		keep_below_default		= parse_bool('Keep_below', False)		# keep the container below other windows
		show_buttons_default	= parse_bool('draw_buttons', True)		# draw close and property buttons
		show_ball_default		= parse_bool('Show_ball', False)		# show the applet inside a ball
		enable_physics_default	= parse_bool('Enable_bouncing', False)	# enable the physics engine
		self.compositing_enabled = not parse_bool('Disable_compositing', True) # composite the applet onto the container (this is still a bit buggy)
		
		# used for animations
		self.opacity = 0
		self.fade_timeout_id = None
		self.begin_fade_timeout_id = None
		
		# used for bouncing ball
		self.dragged		= False # true when the window is down and the window is being dragged by the mouse
		self.mouse_over 	= False	# true if the mouse is currently over the window
		self.mouse_down		= False # true when the mouse button is down
		self.moving			= False	# true when the ball is in motion and is NOT being dragged
		self.vx 			= 0		# velocity along the x axis
		self.vy 			= 0		# velocity along the y axis
		
		# screen edges
		# TODO: update if the screen changes
		self.max_x = gtk.gdk.screen_width()
		self.max_y = gtk.gdk.screen_height()
		
		# shape bitmap and related attributes
		self.__shape_bitmap			= None	# A bitmap used to cache the window shape
		self.__shape_bitmap_width	= 0		# The bitmap's width
		self.__shape_bitmap_height	= 0		# The bitmap's height
		
		# other
		self.grabbed			= False # True when the mouse has been grabbed
		self.menu				= None	# the right click menu
		self.disable_updates	= False	# to temporarily avoid refresh/reshape (e.g. if the container is offscreen)
		self.show_buttons 		= True	# show floating buttons
		
		# Add display related settings
		self.add_options_group('Display', 
			_('The basic settings for this container.'))			
		
		self.add_option(FloatOption('Display', 'scale', 
			1.0, _('Scale'), _('The scale-factor of this Screenlet.'), 
			min=0.1, max=10.0, digits=2, increment=0.1))
			
		self.add_option(IntOption('Display', 'x',
			100, _('X-Position'), _('The X-position of this Screenlet.'), 
			min=0, max=gtk.gdk.screen_width()-int(width*self.scale), callback=self.move))
		
		self.add_option(IntOption('Display', 'y',
			100, _('Y-Position'), _('The Y-position of this Screenlet.'), 
			min=0, max=gtk.gdk.screen_height()-int(height*self.scale), callback=self.move))
		
		self.add_option(FloatOption('Display', 'opacity_max', 
			1.0, _('Opacity'), _('The opacity of the Screenlet window.'), 
			min=0.1, max=1.0, digits=2, increment=0.1))
		
		self.add_option(BoolOption('Display', 'is_sticky', 
			is_sticky_default, _('Stick to Desktop'), 
			_('Show this Screenlet on all workspaces.')))
		
		self.add_option(BoolOption('Display', 'is_widget', 
			is_widget_default, _('Treat as Widget'), 
			_('Show this Screenlet on the Widget Layer. (Requires Compiz.)')))
		
		self.add_option(BoolOption('Display', 'lock_position', 
			lock_position_default, _('Lock position'), 
			_('Stop the screenlet from being moved.')))
		
		self.add_option(BoolOption('Display', 'keep_above', 
			keep_above_default, _('Keep above'), 
			_('Keep this Screenlet above other windows.')))
		
		self.add_option(BoolOption('Display', 'keep_below', 
			keep_below_default, _('Keep below'), 
			_('Keep this Screenlet below other windows.')))
		
		self.add_option(BoolOption('Display', 'skip_pager', 
			True, _('Skip Pager'), 
			_('Set this Screenlet to show/hide in pagers.')))
		
		self.add_option(BoolOption('Display', 'skip_taskbar', 
			True, _('Skip Taskbar'), 
			_('Set this Screenlet to show/hide in taskbars.')))
		
		# Add style related options
		self.add_options_group('Display Style',
			_('Options related to the container\'s style'))
		
		#self.add_option(BoolOption('Display Style', 'show_buttons',
		#	show_buttons_default, _('Show buttons'),
		#	_('Show floating "close" and "properties" buttons.')))
		
		self.add_option(BoolOption('Display Style', 'show_background',
			False, _('Show background'), _('Draw a background behind the applet.')))
		
		# Add physics related options
		self.add_options_group('Physics Engine', 
			_('Options related to the physics engine.'))
			
		self.add_option(BoolOption('Physics Engine', 'enable_physics',
			enable_physics_default, _('Enable physics'), _('Enable the physics engine')))
		
		self.add_option(StringOption('Physics Engine', 'gravity_level',
			'Medium', _('Gravity'), _('Gravity'), choices=['High', 'Medium', 'Low']))
		
		self.add_option(StringOption('Physics Engine', 'bounce_level',
			'Medium', _('Bounce'), _('Bounce'), choices=['High', 'Medium', 'Low']))
		
		self.add_option(StringOption('Physics Engine', 'friction_level',
			'Medium', _('Friction'), _('Friction'), choices=['High', 'Medium', 'Low']))
		
		#self.add_option(IntOption('Physics Engine', 'drag_time', 
		#	10, _('Drag Time'), _('Time in milliseconds to calculate changes in the applet\'s direction and speed.'),
		#	hidden=True))
		
		#self.add_option(IntOption('Physics Engine', 'motion_time', 
		#	10, _('Motion Time'), _('Time in milliseconds to move the applet.'),
		#	hidden=True))
		
		#self.add_option(IntOption('Physics Engine', 'redraw_time', 
		#	10, _('Redraw Time'), _('Time in milliseconds to redraw the applet during animations.'),
		#	hidden=True))
		
		self.drag_time = 10
		self.motion_time = 10
		self.redraw_time = 10
		
		# set the physics attributes to their initial values (see __setattr__)
		# TODO: Replace with something that's not just a quick hack
		self.gravity_level = self.gravity_level
		self.bounce_level = self.bounce_level
		self.friction_level = self.friction_level
		
		# Create the window
		self.window = gtk.Window()
		
		# Hide the window's background
		self.window.set_app_paintable(True)
		screenlets.alpha_screen_changed(self.window)
		
		# Hide the window's border and set the border size
		# so that we can draw our own window borders if necessary
		self.window.set_decorated(False)
		if self.show_buttons:
			self.window.set_border_width(self.BUTTONS_WIDTH)
			self.window.resize(int(self.width*self.scale)+32, int(self.height*self.scale)+32)
		else:
			self.window.resize(int(self.width*self.scale), int(self.height*self.scale))
		
		# set window settings
		self.window.move(self.x, self.y)
		self.window.set_keep_above(self.keep_above)
		self.window.set_keep_below(self.keep_below)
		self.window.set_skip_taskbar_hint(self.skip_taskbar)
		self.window.set_skip_pager_hint(self.skip_pager)
		if self.is_sticky:
			self.window.stick()
			
		# set type hint
		if str(screenlets.sensors.sys_get_window_manager()).lower() == 'kwin':
			log.warning("You are running kwin. Screenlets doesn't have full compatibility with this window manager.")
		else:
			self.window.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_TOOLBAR)	
		
		# Connect the window's events
		self.window.set_events(gtk.gdk.ALL_EVENTS_MASK)
		self.window.connect("expose-event", self.expose)
		self.window.connect("realize", self.realize)
		self.window.connect("button-press-event", self.button_press_event)
		self.window.connect("button-release-event", self.button_release_event)
		self.window.connect("screen-changed", screenlets.alpha_screen_changed)
		self.window.connect("composited-changed", self.composite_changed)
		self.window.connect("enter-notify-event", self.enter_notify_event)
		self.window.connect("leave-notify-event", self.leave_notify_event)
		self.window.connect("focus-in-event", self.focus_in_event)
		self.window.connect("motion-notify-event", self.motion_notify_event)
		self.window.connect("configure-event", self.configure_event)
		self.window.connect("grab-broken-event", self.grab_broken_event)
		
		# Create the socket
		self.socket = gtk.Socket()
		self.socket.set_events(gtk.gdk.ALL_EVENTS_MASK)
		self.socket.connect("realize", self.realize, plug_id)
		self.socket.connect("plug-removed", self.plug_removed)
		screenlets.alpha_screen_changed(self.socket)
		self.window.add(self.socket)

		# Show the window
		self.window.show_all()
		
		if self.is_widget:
			self.set_is_widget(True)
	
	def __setattr__ (self, name, value):
		super(ToplevelContainer, self).__setattr__(name, value)
		# TODO: Replace this with something that isn't as hackish
		if name in ["gravity_level", "bounce_level", "friction_level"]:
			val = getattr(self, value.upper() + "_" + name[:-6].upper())
			setattr(self, name[:-6], val)
		
	def create_menu (self):
		"""Create the right click menu."""
		# load the menu from a gtk.Builder xml file
		self.builder = gtk.Builder()
		self.builder.add_from_file("%s/share/melange/gtk_builder_gui.xml" % screenlets.INSTALL_PREFIX)
		
		# initialize all of the menuitems to their default values
		for id in ["lock_position", "is_sticky", "is_widget", "keep_above", "keep_below", "keep_below"]:
			item = self.builder.get_object(id)
			item.set_active(getattr(self, id))
		scale_item = self.builder.get_object("scale%s" % int(self.scale * 100))
		
		if scale_item is None:
			log.warning("Unknown scale")
			self.scale = 1.0
			scale_item = self.builder.get_object("scale100")
		scale_item.set_active(True)
		
		self.builder.connect_signals(self)
		self.menu = self.builder.get_object("menu1")
	
	# ----------------------------------------------------------------------
	# Drawing Functions
	# ----------------------------------------------------------------------
	
	def composite_socket (self, ctx, alpha=True):
		"""Composites the container's socket onto the context"""
		if self.socket is None or self.socket.window is None:
			return
		
		ctx.save()
		
		if self.show_background or self.show_buttons:
			ctx.translate(self.BUTTONS_WIDTH, self.BUTTONS_WIDTH)
		
		# workaround for python bug #491256 (http://bugzilla.gnome.org/show_bug.cgi?id=491256)
		ctx.set_source_surface(self.socket.window.cairo_create().get_target(), 0, 0)
		
		ctx.set_operator(cairo.OPERATOR_OVER)
		if alpha:
			ctx.paint()
		else:
			ctx.paint_with_alpha(1)
		ctx.restore()
	
	def draw_buttons (self, ctx, alpha=True):
		"""Draw Melange's floating buttons onto a cairo context."""
		# draw the close icon
		ctx.save()
		ctx.translate(self.window.allocation.width-self.BUTTONS_WIDTH,0)
		ctx.set_source_pixbuf(self.server.close_icon, 0, 0)
		if alpha:
			ctx.paint_with_alpha(self.opacity)
		else:
			ctx.paint()
		ctx.restore()
		
		# draw the properties icon
		ctx.save()
		ctx.translate(self.window.allocation.width-self.BUTTONS_WIDTH*2,0)
		ctx.set_source_pixbuf(self.server.pref_icon, 0, 0)
		if alpha:
			ctx.paint_with_alpha(self.opacity)
		else:
			ctx.paint()
		ctx.restore()
	
	def draw_background (self, ctx, transparency):
		"""Draw a background on a cairo context."""
		ctx.save()
		if transparency != -1:
			ctx.set_source_rgba(1, 1, 1, transparency)
		else:
			ctx.set_source_rgb(1,1,1)
		self.draw_rounded_rectangle(ctx, 0, 0, 20,
			self.window.allocation.width, self.window.allocation.height)
		ctx.restore()
	
	def fade_timeout (self):
		"""Timeout called to fade in/out the buttons.
		direction: -1 to fade out and 1 to fade in."""
		# update the transparency level
		if self.mouse_over and self.opacity < 1.0:
			self.opacity += 0.07
			if self.opacity > 1.0:
				self.opacity = 1.0
		elif self.opacity > 0:
			self.opacity -= 0.05
			if self.opacity < 0.0:
				self.opacity = 0.0
		
		# redraw the canvas
		self.redraw_canvas_area(self.window.allocation.width -
			self.BUTTONS_WIDTH * 2, 0, self.BUTTONS_WIDTH * 2,
			self.BUTTONS_WIDTH)
		
		# if the fade effect is complete then clean up and cancel
		if self.opacity >= 1.0 or self.opacity <= 0.0:
			# if we finished fading out, remove the buttons from the input shape
			# they'll be added back into the shape next time enter_notify_event is called 
			if not self.mouse_over:
				self.update_shape()
			self.fade_timeout_id = None
			return False
				
		# Important: return True so that this function will be called again
		return True
	
	# ----------------------------------------------------------------------
	# Physics engine and window managment
	# ----------------------------------------------------------------------
		
	def change_window_settings (self, widget):
		"""Callback for the window settings menu.
		data is only used for the keep_above and keep_below menus to disable each another."""
		id = widget.get_name()
		log.debug("%s was pressed" % id)
		if id=="lock_position":
			self.lock_position = widget.get_active()
		elif id=="is_sticky":
			self.is_sticky = widget.get_active()
			if self.is_sticky:
				self.window.stick()
			else:
				self.window.unstick()
		elif id=="is_widget":
			self.is_widget = widget.get_active()
			if self.window.window:
				self.set_is_widget(self.is_widget)
		elif id=="keep_above":
			self.keep_above = widget.get_active()
			if self.keep_above and self.keep_below: 
				self.keep_below = False
				self.builder.get_object("keep_below").set_active(False)
			self.window.set_keep_above(self.keep_above)
		elif id=="keep_below":
			self.keep_below = widget.get_active()
			if self.keep_above and self.keep_below: 
				self.keep_above = False
				self.builder.get_object("keep_above").set_active(False)
			self.window.set_keep_below(self.keep_below)
		return False
	
	def drag_timeout (self):
		"""Timeout called when the applet is being dragged."""
		# todo: handle changes in direction by (re)setting begin values
		pass
	
	def motion_timeout (self):
		"""Timeout called when the applet is in motion."""
		log.debug("In motion")
		if not self.moving:
			return False
		# apply gravity to the current speed
		if self.y < self.max_y:
			self.vy += self.gravity
		self.__dict__['x'] += self.vx # calculate new x position
		self.__dict__['y'] += self.vy # calculate new y position
		
		# right screen edge
		if self.x + self.window.allocation.width > self.max_x:
			self.vx *= self.bounce
			self.__dict__['x'] = self.max_x - self.window.allocation.width
		# left screen edge
		elif self.x < 0:
			self.vx *= self.bounce
			self.__dict__['x'] = 0
		# bottom screen edge
		if self.y + self.window.allocation.height > self.max_y:
			self.vy *= self.bounce
			self.__dict__['y'] = self.max_y - self.window.allocation.height
		# top screen edge
		elif self.y < 0:
			self.vy *= self.bounce
			self.__dict__['y'] = 0
		
		# if the x velocity is too small just make it 0
		if abs(self.vx) < 1:
			self.vx = 0
		# if the y velocity is too small just make it 0
		if abs(self.vy) < 1:
			self.vy = 0
		
		# if there's no movement in any direction stop moving
		if self.y + self.window.allocation.height == self.max_y and self.vy == 0:
			self.vx *= self.friction
			if self.vx == 0:
				self.moving = False # the timeout will automatically cancel itself
		# finally, move
		self.window.move(int(self.x), int(self.y)) # move to the new location
		return True
	
	def move (self, x=-1, y=-1):
		"""Moves the main window to x,y.
		x: x location on the screen or -1 to use the current location
		y: y location on the screen or -1 to use the current location"""
		if x == -1: x = self.x
		if y == -1: y = self.y
		self.window.move(x, y)
	
	def reshow(self):
		"""Reshow the window"""
		self.window.show()
		self.window.present()
		self.set_is_widget(self.is_widget)
		self.window.set_keep_above(self.keep_above)
		self.window.set_keep_below(self.keep_below)
		if self.window.window:
			self.window.window.set_skip_taskbar_hint(self.skip_taskbar)
			self.window.window.set_skip_pager_hint(self.skip_pager)
		self.has_focus = False
	
	def set_is_widget (self, value):
		"""Set this window to be treated as a Widget.
		Note: This is currently only supported by
		compiz using the widget plugin."""
		if value==True:
			self.window.window.property_change("_COMPIZ_WIDGET", 
				gtk.gdk.SELECTION_TYPE_WINDOW,
				32, gtk.gdk.PROP_MODE_REPLACE, (True,))
		else:
			self.window.window.property_delete("_COMPIZ_WIDGET")
	
	def set_size (self, width, height, size_type):
		"""
		Sets the applets size and/or size type.
		Note: size_type is currently ignored and assumed to be "scalable"
		"""
		# Set the size
		if width != -1: self.width = width
		if height != -1: self.height = height
		if size_type != "": self.size_type = size_type
		
		# Resize the window
		if self.show_buttons:
			self.window.resize(int(self.width*self.scale)+32, int(self.height*self.scale)+32)
		else:
			self.window.resize(int(self.width*self.scale), int(self.height*self.scale))
			
	# ----------------------------------------------------------------------
	# Event handlers for GTK-events
	# ----------------------------------------------------------------------
	
	def button_press_event (self, widget, event):
		"""Event handler for the window's and socket's button-press-events."""
		if widget is not None:
			w = widget.allocation.width
			# check if the user clicked on the quit button
			if self.show_buttons and event.x >= w - self.BUTTONS_WIDTH and event.y <= self.BUTTONS_WIDTH:
				self.menuitem_callback(None, 'quit')
				return
			# check if the user clicked on the settings button OR right clicked anywhere
			elif self.show_buttons and event.x >= w - self.BUTTONS_WIDTH*2 and event.y <= self.BUTTONS_WIDTH or event.button == 3: 
				self.menu.popup(None, None, None, 0, int(event.time))
				return
		# if the applet "forwarded" us this event then grab the 
		# pointer so that we'll recieve button-release-events
		else:
			log.debug("Grabbing pointer")
			grab_status = gtk.gdk.pointer_grab(self.window.window,
				False, gtk.gdk.POINTER_MOTION_MASK |
				gtk.gdk.POINTER_MOTION_HINT_MASK |
				gtk.gdk.BUTTON_MOTION_MASK |
				gtk.gdk.BUTTON_PRESS_MASK |
				gtk.gdk.BUTTON_RELEASE_MASK |
				gtk.gdk.ENTER_NOTIFY_MASK |
				gtk.gdk.LEAVE_NOTIFY_MASK,
				None, None, 0L)
			log.debug("Grab status: %s" % grab_status)
			self.grabbed = True
			
			# Start dragging the window, if not locked.
			if not self.lock_position:
				# Get the absolute position of the pointer.
				# This must be done the hacky way, because the event was forwarded
				# through the dbus server.
				pointer = self.window.get_pointer()
				position = self.window.get_position()
				x, y = [i + j for i, j in zip(pointer, position)]
				self.window.begin_move_drag(0, x, y, 0L)
				self.dragged = True
			
			self.mouse_down = True
	
	def button_release_event (self, widget, event):
		"""Event handler for button releases. Calls gtk.gdk.pointer_ungrab(),
		forwards the event to the applet, and sets self.mouse_down to False"""
		if self.grabbed:
			# ungrab the mouse if we grabbed it in self.button_press_event
			log.debug("button-release-event: ungrabbing pointer")
			gtk.gdk.pointer_ungrab()
			
			# forward the applet the button-press-event that it would have received
			# had we not grabbed the mouse
			self.server.button_release_event(self.id, event.x, event.y,
			    event.state, event.button, event.x_root, event.y_root)
		
		self.mouse_down = False
		self.dragged = False

	def plug_removed (self, widget):
		"""Signal handler called when the plug is removed."""
		if self.menu:
			self.menu.destroy()
		self.window.destroy()
		self.server.remove_applet(self)
		del self
	
	def configure_event (self, widget, event):
		if event.x != self.x or event.y or event.y != self.y:
			self.__dict__['x'] = event.x
			self.__dict__['y'] = event.y
			self.backend.save_option(self.id, 'x', str(event.x))
			self.backend.save_option(self.id, 'y', str(event.y))
			
			# if the window is stickied, make sure that the user didn't drag it offscreen
			if self.is_sticky:
				if self.x + self.window.allocation.width - self.BUTTONS_WIDTH <= 20:
					self.x = gtk.gdk.screen_width() - (self.x % gtk.gdk.screen_width())
				elif self.x >= gtk.gdk.screen_width() - 20:
					self.x = self.x % gtk.gdk.screen_width()
				
				if self.y + self.window.allocation.height - self.BUTTONS_WIDTH <= 20:
					self.y = 0
				elif self.y >= gtk.gdk.screen_height() - self.BUTTONS_WIDTH - 20:
					self.y = gtk.gdk.screen_height() - self.window.allocation.height
				
		return False
	
	def enter_notify_event (self, widget, event):
		if self.mouse_over == False:
			self.mouse_over = True
			if self.fade_timeout_id is None:
				# Note: There are two timers involved in a fade:
				#
				# self.fade_timeout: Called a few times every second to actually fade and redraw the buttons. This will 
				#	automatically reverse direction (fade in/out) if the mouse moves in/out of the window and will also
				#	cancel itself when it's finished.
				#	
				#	This timeout is running when self.fade_timeout is not None.
				#
				# begin_fade (defined below): Instead of calling self.fade_timeout directly, begin_fade is used to call 
				#	self.fade_timeout a short time _after_ we recieve a notify event. This wait is done to prevent the
				# 	buttons being shown if the user quickly moved the mouse in and then out of the window.
				#	In that case, we'll cancel the wait so that the fade effect is never started.
				#
				#	This timeout is running when self.begin_fade_effect_id is not None.
				
				# If we began a wait to begin fading the buttons, and the mouse is now back where is started,
				# then cancel the wait and DON'T fade the buttons at all
				if self.begin_fade_timeout_id is not None:
					gobject.source_remove(self.begin_fade_timeout_id)
					self.begin_fade_timeout_id = None
				
				# Otherwise, just set a timeout that will call the function begin_fade after a wait.
				# This may be cancelled if the mouse quickly moves back in/out of the window. (See the if statement above)
				else:
					# Define the begin_fade function.
					def begin_fade ():
						# Add the buttons to the window's input shape so that the user can click on them.
						# The buttons will be removed from the shape in the function self.fade_timeout at the end of a fade-out. 
						self.update_shape()
						
						# start the timeout to begin fading the buttons
						self.fade_timeout_id = gobject.timeout_add(self.FADE_TIME, self.fade_timeout)
						
						# now that we're done, set this function's timeout id to None so that it'll work next time around
						self.begin_fade_timeout_id = None
						
						# important: return False to stop this timeout from being called again
						return False
					
					self.begin_fade_timeout_id = gobject.timeout_add(800, begin_fade)
	
	def focus_in_event (self, widget, event):
		"""Called when the applet receives focus.
		Note: this is also called when the user stops dragging the window."""
		# if the user was in the middle of dragging the window and just let go of the mouse
		if self.mouse_down:
			self.dragged = False
			self.mouse_down = False
			
			# if we grabbed the pointer from the applet then let it go
			if self.grabbed:
				log.debug("Ungrabbing pointer on focus-in-event")
				gtk.gdk.pointer_ungrab()
			
			# if the physics engine is enabled then calculate the applet's speed and set it in motion
			if self.enable_physics:
				self.moving = True
				
				# get the current time and position
				now = gobject.get_current_time()
				self.x, self.y = self.window.get_position()
				
				# calculate the velocity along the x and y axis
				self.vx = ((self.x - self.begin_x) / ((now - self.begin_time) * 100)) * self.BOOST
				self.vy = ((self.y - self.begin_y) / ((now - self.begin_time) * 100)) * self.BOOST
				
				# start the timeout function
				self.move_timeout = gobject.timeout_add(self.motion_time, self.motion_timeout)
			
			self.dragged = False

	def grab_broken_event (self, widget, event):
		"""Called if we grabbed the pointer and then the grab was broken."""
		log.debug("Grab broken")

	def leave_notify_event (self, widget, event):
		"""Called when the mouse moves away from the window.
		"""
		if not self.show_background or \
			(event is not None and (event.x <= 0 or event.x >= self.window.allocation.width \
				or event.y <= 0 or event.y >= self.window.allocation.height)):
			
			if self.mouse_over:
				self.mouse_over = False
				
				# see comments in self.enter_notify_event above for an explanation
				if self.fade_timeout_id is None:
					if self.begin_fade_timeout_id is not None:
						gobject.source_remove(self.begin_fade_timeout_id)
						self.begin_fade_timeout_id = None
					else:
						def begin_fade ():
							self.fade_timeout_id = gobject.timeout_add(self.FADE_TIME, self.fade_timeout)
							self.begin_fade_timeout_id = None
							return False
						
						self.begin_fade_timeout_id = gobject.timeout_add(800, begin_fade)

	def motion_notify_event (self, widget, event):
		"""Callback for mouse movement over the applet."""
		# if the mouse is pressed, we should let the user drag the window across the screen
		if self.mouse_down and not self.lock_position and not self.dragged:
			self.dragged = True
			
			# if the physics engine is enabled
			if self.enable_physics:
				self.moving = False # stop bouncing if we were already in motion
				self.begin_x, self.begin_y = self.window.get_position()
				self.begin_time = gobject.get_current_time()
				self.update_speed_timeout = gobject.timeout_add(self.drag_time, self.drag_timeout)

	def menuitem_callback (self, widget, id=None):
		if id is None:
			id = widget.get_name()
		if id == "quit":
			self.window.destroy()
			self.server.remove_applet(self)
		elif id in ("info", "about", "settings", "options", "properties", "display settings"):
			# show settings dialog
			self.show_options_dialog()
		elif id.startswith('scale'):
			#self.applet.scale = float(id[6:])
			# necessary?
			self.scale = float(id[5:]) / 100
			if self.window:
				width = int(int(self.width*self.scale))
				height = int(int(self.height*self.scale))
				if self.show_buttons:
					self.window.resize(width+32, height+32)
				else:
					self.window.resize(width, height)
				log.info("Adding update_shape timeout")
				gobject.timeout_add(300, self.update_shape)
		elif id[:8] == "setting:":
			# set a boolean option to the opposite state
			try:
				if type(self.__dict__[id[8:]]) == bool:
					self.__dict__[id[8:]] = not self.__dict__[id[8:]]	# UNSAFE!!
			except:
				log.error("Cannot set missing or non-boolean value '%s'" % id[8:])
		return False
	
	def realize (self, widget, plug_id=None):
		"""Realize handler for both the socket's and window's realize events"""
		# if the socket realized add the plug
		if plug_id:
			#TODO check for self.socket.window.supports_composite
			#see here for the details: http://library.gnome.org/devel/pygtk/2.12/class-gdkdisplay.html
			if self.compositing_enabled:
				self.socket.window.set_composited(True)
			self.socket.add_id(long(plug_id))
			self.create_menu()
		# if the window realized take care of other stuff
		else:
			if self.window.window: # necessary?
				self.window.window.set_back_pixmap(None, False)	# needed?
				#self.style = self.window.get_style()
				#self.update_shape()
	
	def expose (self, widget, event):
		"""Event handler for the window's expose event.""" 
		# if updates are disabled (e.g. the container is hidden) return
		if self.disable_updates:
			return
		
		log.debug("expose event")
		
		# create the cairo context, save it, and clear it
		ctx = widget.window.cairo_create()
		
		# if we're in the middle of fading the buttons, only redraw them
		if event.area.x == widget.allocation.width - \
		self.BUTTONS_WIDTH * 2 and event.area.y == 0 \
		and event.area.width == self.BUTTONS_WIDTH * 2 \
		and event.area.height == self.BUTTONS_WIDTH:
			Drawing.clear_cairo_context(ctx, widget.allocation.width -
				self.BUTTONS_WIDTH * 2, 0, self.BUTTONS_WIDTH * 2,
				self.BUTTONS_WIDTH)
			self.draw_buttons(ctx)
			return
				
		Drawing.clear_cairo_context(ctx)
		
		# redraw the background
		if self.show_background and self.opacity > 0.0:
			if self.opacity > 0.3:
				transparency = 0.3
			else:
				transparency = self.opacity
			self.draw_background(ctx, transparency)
			
		# redraw the icons
		if self.show_buttons and self.opacity > 0.0:
			self.draw_buttons(ctx)
		
		# composite the applet onto the context
		if self.compositing_enabled:
			self.composite_socket(ctx)
	
	def composite_changed (self, widget):
		"""Signal handler called when the composited state changes.
		TODO: Fix up the broken bits."""
		log.info(_('Compositing method changed to %s') % self.window.is_composited())
		self.window.hide() # hide the window and show it again so that it can convert to the right composited state
		self.window.show()
		if self.window:
			if self.is_sticky:
				self.window.stick()
			else:
				self.window.unstick()
			self.window.set_keep_above(self.keep_above)
			self.window.set_keep_below(self.keep_below)
				
	# Other
	
	def redraw_canvas (self):
		"""Redraw the entire Screenlet's window area."""			
		# if updates are disabled, just exit
		if self.disable_updates or self.window.window is None:
			return
		
		x, y, w, h = widget.get_allocation()
		rect = gtk.gdk.Rectangle(x, y, w, h)
		self.window.window.invalidate_rect(rect, True)
		self.window.window.process_updates(True)
	
	def redraw_canvas_area (self, x, y, width, height):
		"""Redraw the given Rectangle (x, y, width, height) within the 
		current Screenlet's window."""
		# if updates are disabled, just exit
		if self.disable_updates or self.window.window is None:
			return
		
		rect = gtk.gdk.Rectangle(x, y, width, height)
		self.window.window.invalidate_rect(rect, True)
		self.window.window.process_updates(True)
	
	def update_shape(self):
		"""Update the window's input shape."""
		if self.window is None or self.window.window is None:
			return
		
		w = self.window.allocation.width
		h = self.window.allocation.height
		
		# if 0 return to avoid crashing
		if w==0 or h==0: return False
		# if size changed, recreate shape bitmap
		if w != self.__shape_bitmap_width or h != self.__shape_bitmap_height:
			self.__shape_bitmap = screenlets.create_empty_bitmap(w, h)
			self.__shape_bitmap_width = w
			self.__shape_bitmap_height = h
			
		# create context
		ctx = self.__shape_bitmap.cairo_create()
		Drawing.clear_cairo_context(ctx)
		
		# draw the background only because it covers everything else
		if self.show_background and self.opacity > 0.0:
			self.draw_background(ctx, -1)
		else:
			# draw icons
			if self.show_buttons and self.mouse_over:
				self.draw_buttons(ctx, False)
			# composite the applet onto the context
			#if self.direct_shaping_enabled:
			self.composite_socket(ctx, False)
		
		# shape the window acording if the window is composited  or not
		if self.window.is_composited():
			log.debug("Updating input shape")
			self.window.window.input_shape_combine_mask(self.__shape_bitmap, 0, 0)
			#self.window.window.merge_child_input_shapes()
		else:
			log.debug("Updating window shape")
			self.window.window.shape_combine_mask(self.__shape_bitmap, 0, 0)
			#self.window.window.merge_child_shapes()
	
	def show_options_dialog (self):
		"""Show the EditableSettingsDialog for this ScreenletContainer.
		TODO: Merge this window into the applet's settings window."""
		se = OptionsDialog(490, 450)
		
		img = gtk.Image()
		img.set_from_stock(gtk.STOCK_PREFERENCES, 5)
			
		se.set_title(self.server.get_server_name())
		se.set_info(self.server.get_server_name(), "Melange is an applet server that displays applets \
			in floating windows that can be dragged around the screen.", "(c) Natan Yellin",
			version=("v%s" % self.server.get_server_version()), icon=img)
		se.show_options_for_object(self)
		resp = se.run()
		if resp == gtk.RESPONSE_REJECT:	# TODO!!!!!
			se.reset_to_defaults()
		se.destroy()
