# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Options Module- (c) 2008 Natan Yellin (Aantn) <aantny@gmail.com>
# See AUTHORS for a list of other authors and copyright holders. 
#
# INFO:
# - A dynamic system that allows developers to easily create objects
# that can load/save their settings from file and/or allow users to edit
# those settings with an automatically generated GUI.
#
# TODO:
# - Ensure that all of Option's attributes do what they should and are well documented.
#
# Other Possible Improvements:
# - option-widgets for all option-types (e.g. ListOptionWidget, ColorOptionWidget)
# - OptionGroup-class instead of (or behind) add_options_group
# - DateOption
# - FileOption needs filter/limit-attribute
# - allow options to disable/enable other options
# - support for EditableOptions-subclasses as options
# - separate OptionEditorWidget from Editor-Dialog
# - place ui-code into screenlets.options.ui-module
#

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

import os		
import xml.dom.minidom
from xml.dom.minidom import Node

import gtk
import gettext
import gnomekeyring
import gobject

import screenlets
from . import utils

# translation stuff
gettext.textdomain('screenlets')
gettext.bindtextdomain('screenlets', screenlets.INSTALL_PREFIX +  '/share/locale')

def _(s):
	return gettext.gettext(s)


# log
log = screenlets.logger.get_default_logger()
# -----------------------------------------------------------------------
# Option-classes and subclasses
# -----------------------------------------------------------------------

class Option(object):
	"""An Option stores information about a certain object-attribute.
	Every option contains a name, default value, label, etc. and
	knows how to display itself in a gui."""
	
	def __init__ (self, group, name, default, label='', desc='',  
		enabled=True, hidden=False, callback=None, always_trigger_callback=True,
		realtime=True, protected=False):
		"""Creates a new Option with the given information.
		
		group: Name of an options-group used to group options into tabs in the GUI. 
		name: Used as a unique identifier.
		default: Default value for this option.
		label: Quick description of this option. This is displayed to the user in the GUI for editing options.
		desc: Detailed description of this option. This is displayed to the user as a tooltip in the GUI for editing options.
		disabled: DOCUMENTATION GOES HERE.
		hidden: DOCUMENTATION GOES HERE.
		callback: Function to be called when this option's value changes.
		always_trigger_callback: If True, the callback will always be called,
		 	even when the option's value is changed from internal code in your
			application. If False, the callback will only be called when the
			option is changed by the user via the automatically generated gui.
		realtime: If False, this option only takes effect once the user clicks an "apply" button.
		protected: DOCUMENTATION GOES HERE.
		"""
		super(Option, self).__init__()
		# see above for an explanation of these attributes 
		self.name 		= name
		if label == '':
			self.label = self.name
		else:
			self.label = label
		if desc == '':
			self.desc = self.label
		else:
			self.desc = desc
		self.default	= default
		self.enabled	= enabled
		self.hidden		= hidden
		self.group		= group
		self.callback	= callback
		self.always_trigger_callback = always_trigger_callback
		self.realtime = realtime
		self.protected = protected
	
	def get_widget (self, value, editable_options):
		widget = gtk.Entry()
		log.error(_("This option class doesn't support the options gui:") + str(self.__class__))
	
	def get_value_from_widget (self, widget):
		print _("OptionsDialog: Unknown option type: %s") % str(self.__class__)
		return None
	
	def on_import (self, strvalue):
		"""Called when an option gets imported from a string. 
		This function MUST return the string-value converted to the required 
		type!"""
		return strvalue.replace("\\n", "\n")
	
	def on_export (self, value):
		"""Called when an option gets exported to a string.
		This method MUST return the value converted to a string!"""
		return str(value).replace("\n", "\\n")


class AccountOption (Option):
	"""An Option-type for username/password combos. Stores the password in
	the gnome-keyring (if available) and only saves username and auth_token
	through the screenlets-backend.
	TODO:
	- not create new token for any change (use "set" instead of "create" if
	  the given item already exists)
	- use usual storage if no keyring is available but output warning
	- on_delete-function for removing the data from keyring when the
	  Screenlet holding the option gets deleted"""
	
	def __init__ (self, group, name, default, label, desc, **keyword_args):
		Option.__init__ (self, group, name, default, label, desc, 
			protected=True, **keyword_args)
		# check for availability of keyring
		if not gnomekeyring.is_available():
			raise Exception(_('GnomeKeyring is not available!'))	# TEMP!!!
		# THIS IS A WORKAROUND FOR A BUG IN KEYRING (usually we would use
		# gnomekeyring.get_default_keyring_sync() here):
		# find first available keyring
		self.keyring_list = gnomekeyring.list_keyring_names_sync()
		if len(self.keyring_list) == 0:
			raise Exception(_('No keyrings found. Please create one first!'))
		else:
			# we prefer the default keyring
			try:
				self.keyring = gnomekeyring.get_default_keyring_sync()
			except:
				log.warning(_("No default keyring found. Storage is not permanent."))
				self.keyring = self.keyring_list[0]
	
	def get_widget (self, value, editable_options):
		widget = gtk.HBox()
		vb = gtk.VBox()
		input_name = gtk.Entry()
		input_name.set_text(value[0])
		input_name.show()
		input_pass = gtk.Entry()
		input_pass.set_visibility(False)	# password
		input_pass.set_text(value[1])
		input_pass.show()
		but = gtk.Button('Apply', gtk.STOCK_APPLY)
		but.show()
		but.connect("clicked", editable_options.apply_options_callback, self, widget)
		vb.add(input_name)
		vb.add(input_pass)
		vb.show()
		editable_options.tooltips.set_tip(but, _('Apply username/password'))
		editable_options.tooltips.set_tip(input_name, _('Enter username here'))
		editable_options.tooltips.set_tip(input_pass, _('Enter password here'))
		widget.add(vb)
		widget.add(but)
		return widget
	
	def get_value_from_widget (self, widget):
		for c in widget.get_children():
			if c.__class__ == gtk.VBox:
				c2 = c.get_children()
				return (c2[0].get_text(), c2[1].get_text())
	
	def on_import (self, strvalue):
		"""Import account info from a string of the format 'username:auth_token', 
		retrieve the password from the storage, and return a tuple containing
		username and password."""
		# split string into username/auth_token
		#data = strvalue.split(':', 1)
		(name, auth_token) = strvalue.split(':', 1)
		if name and auth_token:
			# read pass from storage
			try:
				if self.keyring == self.keyring_list[0]:
					pw = gnomekeyring.item_get_info_sync('session', 
						int(auth_token)).get_secret()
				else:
					pw = gnomekeyring.item_get_info_sync(self.keyring, 
						int(auth_token)).get_secret()
			except Exception, ex:
				log.error(_("Unable to read password from keyring: ") + str(ex))
				pw = ''
			# return
			return (name, pw)
		else:
			raise Exception(_('Illegal value in AccountOption.on_import.'))
	
	def on_export (self, value):
		"""Export the given tuple/list containing a username and a password. The
		function stores the password in the gnomekeyring and returns a
		string in form 'username:auth_token'."""
		# store password in storage
		attribs = dict(name=value[0])
		if self.keyring == self.keyring_list[0]:
			auth_token = gnomekeyring.item_create_sync('session', 
				gnomekeyring.ITEM_GENERIC_SECRET, value[0], attribs, value[1], True) 
		else:
			auth_token = gnomekeyring.item_create_sync(self.keyring, 
				gnomekeyring.ITEM_GENERIC_SECRET, value[0], attribs, value[1], True)
		# build value from username and auth_token
		return value[0] + ':' + str(auth_token)


class BoolOption (Option):
	"""An Option for boolean values."""
	
	def get_widget (self, value, editable_options):
		widget = gtk.CheckButton()
		widget.set_active(value)
		widget.connect("toggled", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		return widget.get_active()
	
	def on_import (self, strvalue):
		if strvalue == "True":
			return True
		return False


class ColorOption (Option):
	"""An Option for colors. Stored as a list with 4 values (r, g, b, a)."""
	
	def get_widget (self, value, editable_options):
		widget = gtk.ColorButton(gtk.gdk.Color(int(value[0]*65535), int(value[1]*65535), int(value[2]*65535)))
		widget.set_use_alpha(True)
		widget.set_alpha(int(value[3]*65535))
		widget.connect("color-set", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		col = widget.get_color()
		al = widget.get_alpha()
		return (col.red/65535.0, col.green/65535.0, 
			col.blue/65535.0, al/65535.0)
	
	def on_import (self, strvalue):
		"""Import (r, g, b, a) from comma-separated string."""
		# strip braces and spaces
		strvalue = strvalue.lstrip('(')
		strvalue = strvalue.rstrip(')')
		strvalue = strvalue.strip()
		# split value on commas
		tmpval = strvalue.split(',')
		outval = []
		for f in tmpval:
			# create list and again remove spaces
			outval.append(float(f.strip()))
		return outval
	
	def on_export (self, value):
		"""Export r, g, b, a to comma-separated string."""
		l = len(value)
		outval = ''
		for i in xrange(l):
			outval += str(value[i])
			if i < l-1:
				outval += ','
		return outval


class DirectoryOption (Option):
	"""An Option-subclass for filename-strings that contain directories."""
	
	def get_widget (self, value, editable_options):
		dlg = gtk.FileChooserDialog(buttons=(gtk.STOCK_CANCEL, 
			gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK),
			action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		widget = gtk.FileChooserButton(dlg)
		widget.set_title(_("Choose Directory"))
		widget.set_filename(value)
		widget.set_size_request(180, 28)
		widget.connect("selection-changed", editable_options.options_callback, self)
		return widget
		
	def get_value_from_widget (self, widget):
		return widget.get_filename()


class FileOption (Option):
	"""An Option-subclass for string-values that contain filenames. Adds
	a patterns-attribute that can contain a list of patterns to be shown
	in the assigned file selection dialog. The show_pixmaps-attribute
	can be set to True to make the filedialog show all image-types 
	supported by gtk.Pixmap. If the directory-attributue is true, the
	dialog will ony allow directories."""

	def __init__ (self, group, name, default, label, desc, 
		patterns=['*'], image=False, directory=False, **keyword_args):
		Option.__init__(self, group, name, default,label, desc, **keyword_args)
		self.patterns	= patterns
		self.image		= image
		self.directory	= False
	
	def get_widget (self, value, editable_options):
		widget = gtk.FileChooserButton(_("Choose File"))
		widget.set_filename(value)
		widget.set_size_request(180, 28)
		widget.connect("selection-changed", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		return widget.get_filename()
		
		
class FontOption (Option):
	"""An Option for fonts (a simple StringOption)."""
	
	def get_widget (self, value, editable_options):
		widget = gtk.FontButton()
		widget.set_font_name(value)
		widget.connect("font-set", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		return widget.get_font_name()


class ImageOption (Option):
	"""An Option-subclass for string-values that contain filenames of
	image-files."""
	
	def get_widget (self, value, editable_options):
		# create entry and button (entry is hidden)
		entry = gtk.Entry()
		entry.set_text(value)
		entry.set_editable(False)
		button = gtk.Button('')
		
		# load preview image
		button.set_image(self._get_preview_image(value))
		button.connect('clicked', self._on_button_click, entry, editable_options)
		
		# create widget
		widget = gtk.HBox()
		widget.add(entry)
		widget.add(button)
		button.show()
		widget.show()
		
		# add tooltips
		editable_options.tooltips.set_tip(widget, self.desc)
		
		return widget
	
	def get_value_from_widget (self, widget):
		return widget.get_filename()
	
	def _get_preview_image (self, filename):
		"""Get a 64 by 64 pixel preview of filename."""
		if filename and os.path.isfile(filename):
			pb = gtk.gdk.pixbuf_new_from_file_at_size(filename, 64, -1)
			if pb is not None:
				img = gtk.Image()
				img.set_from_pixbuf(pb)
				return img
		img = gtk.image_new_from_stock(gtk.STOCK_MISSING_IMAGE, 
			gtk.ICON_SIZE_LARGE_TOOLBAR)
		img.set_size_request(64, 64)
		return img
	
	def _preview_callback(self, widget, dlg, preview):
		"""Called when the image is changed to
		update the preview image."""
		fname = dlg.get_preview_filename()
		if fname and os.path.isfile(fname):
			pb = gtk.gdk.pixbuf_new_from_file_at_size(fname, 150, -1)
			if pb:
				preview.set_from_pixbuf(pb)
				dlg.set_preview_widget_active(True)
			else:
				dlg.set_preview_widget_active(False)
						
	def _on_button_click (self, widget, entry, editable_options):
		"""Callback for button clicks."""
		dlg = gtk.FileChooserDialog(buttons=(gtk.STOCK_CANCEL, 
			gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
		dlg.set_title(_("Choose Image"))
		dlg.set_keep_above(True)
		dlg.set_filename(entry.get_text())
		flt = gtk.FileFilter()
		flt.add_pixbuf_formats()
		dlg.set_filter(flt)
		prev = gtk.Image()
		box = gtk.VBox()
		box.set_size_request(150, -1)
		box.add(prev)
		prev.show()
		
		dlg.set_preview_widget_active(True)
		dlg.connect('selection-changed', self._preview_callback, dlg, prev)
		dlg.set_preview_widget(box)
		
		# run 
		response = dlg.run()
		if response == gtk.RESPONSE_OK:
			entry.set_text(dlg.get_filename())
			widget.set_image(self._get_preview_image(dlg.get_filename()))
			editable_options.options_callback(dlg, self)
		dlg.destroy()


class IntOption (Option):
	"""An Option for values of type number (can be int or float)."""
	
	def __init__ (self, group, name, default, label, desc,  min=0, max=100, 
		increment=1, **keyword_args):
		Option.__init__(self, group, name, default, label, desc, **keyword_args)
		self.min = min
		self.max = max
		self.increment = increment
	
	def get_widget (self, value, editable_options):
		widget = gtk.SpinButton()
		widget.set_increments(self.increment, int(self.max/self.increment))
		widget.set_range(self.min, self.max)
		widget.set_value(value)
		widget.connect("value-changed", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		return int(widget.get_value())
		
	def on_import (self, strvalue):
		"""Called when IntOption gets imported. Converts str to int."""
		# find the right sign (+ or -)
		if strvalue[0]=='-':
			val = int(strvalue[1:]) * -1
		else:
			val = int(strvalue)
		
		# check if the value is within min and max
		if val < self.min:
			val = self.min
		elif val > self.max:
			val = self.max
		
		return val


class FloatOption (IntOption):
	"""An Option for values of type float."""
	
	def __init__ (self, group, name, default, label, desc, min=0.0,
		max=1.0, digits=1, **keyword_args):
		IntOption.__init__(self, group, name, default, label, desc,
			min, max, **keyword_args)
		self.digits = digits
	
	def get_widget (self, value, editable_options):
		widget = gtk.SpinButton()
		widget.set_digits(self.digits)
		widget.set_increments(self.increment, int(self.max/self.increment))
		widget.set_range(self.min, self.max)
		widget.set_value(value)
		widget.connect("value-changed", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		return widget.get_value()
	
	def on_import (self, strvalue):
		"""Called when FloatOption gets imported. Converts str to float."""
		# find the right sign (+ or -)
		if strvalue[0]=='-':
			val = float(strvalue[1:]) * -1
		else:
			val = float(strvalue)
		
		# check if the value is within min and max
		if val < self.min:
			val = self.min
		elif val > self.max:
			val = self.max
			
		return val
		
		
class ListOption (Option):
	"""An Option-type for list of strings."""

	def get_widget (self, value, editable_options):
		entry= gtk.Entry()
		entry.set_editable(False)
		entry.set_text(str(value))
		entry.show()
		img = gtk.Image()
		img.set_from_stock(gtk.STOCK_EDIT, 1)
		but = gtk.Button('')
		but.set_image(img)
		
		but.show()
		but.connect("clicked", self.open_listeditor, entry, editable_options)
		editable_options.tooltips.set_tip(but, _('Edit list.'))
		editable_options.tooltips.set_tip(entry, self.desc)
		widget = gtk.HBox()
		widget.add(entry)
		widget.add(but)
		
		return widget
	
	def get_value_from_widget (self, widget):
		# note: widget is a ListOptionsDialog
		return widget.get_list()
		
	def on_import (self, strvalue):
		"""Import python-style list from a string (like [1, 2, 'test'])"""
		lst = eval(strvalue)
		return lst
	
	def on_export (self, value):
		"""Export list as string."""
		return str(value)
	
	def open_listeditor(self, widget, entry, editable_options):
		# open dialog
		dlg = ListOptionDialog()
		# read string from entry and import it through option-class
		# (this is needed to always have an up-to-date value)
		dlg.set_list(self.on_import(entry.get_text()))
		resp = dlg.run()
		if resp == gtk.RESPONSE_OK:
			# set text in entry
			entry.set_text(str(dlg.get_list()))
			# manually call the options-callback
			editable_options.options_callback(dlg, self)
		dlg.destroy()


class StringOption (Option):
	"""An Option for values of type string."""
	
	def __init__ (self, group, name, default, label, desc, 
		choices=None, password=False, **keyword_args):
		Option.__init__(self, group, name, default,label, desc, **keyword_args)
		self.choices	= choices
		self.password	= password
	
	def get_widget (self, value, editable_options):
		if self.choices:
			# if a list of values is defined, show combobox
			widget = gtk.combo_box_new_text()
			p = -1
			i = 0
			for s in self.choices:
				widget.append_text(s)
				if s==value:
					p = i
				i+=1
			widget.set_active(p)
		else:
			widget = gtk.Entry()
			widget.set_text(value)
			# if it is a password, set text to be invisible
			if self.password:
				widget.set_visibility(False)
		widget.connect("changed", editable_options.options_callback, self)
		return widget
	
	def get_value_from_widget (self, widget):
		if self.choices:
			# if default is a list, handle combobox
			return widget.get_active_text()
		else:
			return widget.get_text()


class TimeOption (ColorOption):
	"""An Option-subclass for string-values that contain dates."""
	
	def get_widget (self, value, editable_options):
		box = gtk.HBox()
		
		# create 3 spin buttons for the hour, minute, and second
		i = 0
		for range_max in [23, 59, 59]:
			button = gtk.SpinButton()
			button.set_range(0, range_max)
			button.set_max_length(2)
			button.set_increments(1, 1)
			button.set_numeric(True)
			button.set_value(value[i])
			button.connect('value-changed', editable_options.options_callback, self)
			editable_options.tooltips.set_tip(box, self.desc)
			box.add(button)
			box.add(gtk.Label(':'))
			i += 1
		
		box.show_all()
		return box
	
	def get_value_from_widget (self, widget):
		box = widget.get_parent()
		inputs = box.get_children()
		return (int(inputs[0].get_value()), int(inputs[2].get_value()), 
			int(inputs[4].get_value()))


# -----------------------------------------------------------------------
# EditableOptions-class and needed functions
# -----------------------------------------------------------------------

def create_option_from_node (node, groupname):
	"""Create an Option from an XML-node with option-metadata."""
	otype = node.getAttribute("type")
	oname = node.getAttribute("name")
	ohidden = node.getAttribute("hidden")
	odefault	= None
	oinfo		= ''
	olabel		= ''
	omin		= None
	omax		= None
	oincrement	= 1
	ochoices	= ''
	odigits		= None
	if otype and oname:
		# parse children of option-node and save all useful attributes
		for attr in node.childNodes:
			if attr.nodeType == Node.ELEMENT_NODE:
				if attr.nodeName == 'label':
					olabel = attr.firstChild.nodeValue
				elif attr.nodeName == 'info':
					oinfo = attr.firstChild.nodeValue
				elif attr.nodeName == 'default':
					odefault = attr.firstChild.nodeValue
				elif attr.nodeName == 'min':
					omin = attr.firstChild.nodeValue
				elif attr.nodeName == 'max':
					omax = attr.firstChild.nodeValue
				elif attr.nodeName == 'increment':
					oincrement = attr.firstChild.nodeValue
				elif attr.nodeName == 'choices':
					ochoices = attr.firstChild.nodeValue
				elif attr.nodeName == 'digits':
					odigits = attr.firstChild.nodeValue
		# if we have all needed values, create the Option
		if odefault:
			# create correct classname here
			cls = otype[0].upper() + otype.lower()[1:] + 'Option'
			#print 'Create: ' +cls +' / ' + oname + ' ('+otype+')'
			# and build new instance (we use on_import for setting default val)
			clsobj = OPTIONS_MAPPING[cls]
			opt = clsobj(groupname, oname, None, olabel, oinfo)
			opt.default = opt.on_import(odefault)
			# set values to the correct types
			if cls == 'IntOption':
				if omin:
					opt.min = int(omin)
				if omax:
					opt.max = int(omax)
				if oincrement:
					opt.increment = int(oincrement)
			elif cls == 'FloatOption':
				if odigits:
					opt.digits = int(odigits)
				if omin:
					opt.min = float(omin)
				if omax:
					opt.max = float(omax)
				if oincrement:
					opt.increment = float(oincrement)
			elif cls == 'StringOption':
				if ochoices:
					opt.choices = ochoices
			return opt
	return None
	

class EditableOptions (object):
	"""The EditableOptions class provides a base for classes which need to
	save and load settings from file and/or allow the user to edit those
	options with a GUI."""
	
	saving_enabled = False
	
	def __init__ (self, backend=None, id='', saving_enabled=False):
		"""Constructor.
		backend: The backend to use for saving options. Please see the backend module for more information.
		id: A unique application id to use when saving options. This is different for each configuration stored with the smae backend.
		saving_enabled: Enable saving of options to file. If this is False, options will not be saved to file."""
		
		self.__options__ = []
		self.__options_groups__ = {}
		self.__options_groups_ordered__ = [] # workaround to remember the order of groups
		
		self.__backend = backend
		self.__id = id
		
		# saving enabled
		if saving_enabled and backend is not None and id != '':
			self.saving_enabled = True
		else:
			self.saving_enabled	= False
	
	def __setattr__ (self, name, value):
		# set the attribute in object
		super(EditableOptions, self).__setattr__(name, value)
		
		# get option object
		option = self.get_option_by_name(name)
		if option is not None:
			# save the option in the backend
			if self.saving_enabled:
					self.__backend.save_option(self.__id, option.name, option.on_export(value))
			
			# call the option's callback
			if option.callback is not None and option.always_trigger_callback:
				option.callback()
	
	def add_option (self, option, realtime=True):
		"""Add an editable option to this object. Editable Options can be edited
		and configured using the OptionsDialog. The optional callback-arg can be
		used to set a callback that gets notified when the option changes its 
		value."""
		# if the option is already editable (i.e. initialized), return
		for o in self.__options__:
			if o.name == option.name:
				return False
		
		value = option.default
		# load the options value from the backend so that it doesn't have to be done in session.py
		# for backwards compatibility, we reload the options again in session.py
		if self.__backend:
			str = self.__backend.load_option(self.__id, option.name, option.default)
			if str is not option.default:
				value = option.on_import(str)
		
		self.__dict__[option.name] = value
		
		# set realtime
		if realtime == False:
			log.warning("Please set realtime in Option.__init__ and not in EditableOptions.add_option")
		option.realtime = realtime
		
		# add option to group (output error if group is undefined)
		try:
			self.__options_groups__[option.group]['options'].append(option)
		except:
			log.error(_("Options group is not defined: ") + (str(option.group)))
			return False
		# now add the option
		self.__options__.append(option)
		return True
		
	def add_options_group (self, name, group_info):
		"""Add a new options-group to this Options-object"""
		self.__options_groups__[name] = {'label':name, 
			'info':group_info, 'options':[]}
		self.__options_groups_ordered__.append(name)
	
	def apply_option_overrides (self, overrides):
		"""Apply option overrides."""
		# save the current update status for later and then disable updates
		#disable_updates = self.disable_updates
		#self.disable_updates = True
		for name, value in overrides.items():
			log.debug(_("Override: ") + name)
			o = self.get_option_by_name(name)
			if o and not o.protected:
				# set the option
				setattr(self, name,	o.on_import(value))
			else:
				log.warning(_("Option '%s' not found or protected.") % name)
		# re-enable updates and call redraw/reshape
		#self.disable_updates = disable_updates
		#self.redraw_canvas()
		#self.update_shape()
	
	def enable_option (self, name):
		"""Enable editing of an option in the GUI."""
		o = self.get_option_by_name(name)
		if o is not None:
			o.enabled = True
			return True
		return False
		
	def disable_option (self, name):
		"""Disable editing of an option in the GUI."""
		o = self.get_option_by_name(name)
		if o is not None:
			o.enabled = False
			return True
		return False
	
	def enable_saving (self, enabled=True):
		"""Enable/Disable realtime-saving of options."""
		self.saving_enabled = enabled
	
	def export_options_as_list (self):
		"""Returns all editable options within a list (without groups)
		as key/value tuples."""
		lst = []
		for o in self.__options__:
			lst.append((o.name, getattr(self, o.name)))
		return lst
	
	def get_option_by_name (self, name):
		"""Returns an option in this Options by it's name (or None)."""
		#TODO: This gives wrong results in child classes. Maybe access
		# as class-attribute?"""
		for o in self.__options__:
			if o.name == name:
				return o
		return None
	
	def remove_option (self, name):
		"""Remove an option from this Options."""
		for o in self.__options__:
			if o.name == name:
				del o
				return True
		return True
	
	def add_options_from_file (self, filename):
		"""This function creates options from an XML-file with option-metadata.
		TODO: make this more reusable and place it into module (once the groups
		are own objects)"""
		# create xml document
		try:
			doc = xml.dom.minidom.parse(filename)
		except:
			raise Exception(_('Invalid XML in metadata-file (or file missing): "%s".') % filename)
		# get rootnode
		root = doc.firstChild
		if not root or root.nodeName != 'screenlet':
			raise Exception(_('Missing or invalid rootnode in metadata-file: "%s".') % filename)
		# ok, let's check the nodes: this one should contain option-groups
		groups = []
		for node in root.childNodes:
			# we only want element-nodes
			if node.nodeType == Node.ELEMENT_NODE:
				if node.nodeName != 'group' or not node.hasChildNodes():
					# we only allow groups in the first level (groups need children)
					raise Exception(_('Error in metadata-file "%s" - only <group>-tags allowed in first level. Groups must contain at least one <info>-element.') % filename)
				else:
					# ok, create a new group and parse its elements
					group = {}
					group['name']		= node.getAttribute("name")
					if not group['name']:
						raise Exception(_('No name for group defined in "%s".') % filename)
					group['info']		= ''
					group['options']	= []
					# check all children in group
					for on in node.childNodes:
						if on.nodeType == Node.ELEMENT_NODE:
							if on.nodeName == 'info':
								# info-node? set group-info
								group['info'] = on.firstChild.nodeValue
							elif on.nodeName == 'option':
								# option node? parse option node
								opt = create_option_from_node (on, group['name'])
								# ok? add it to list
								if opt:
									group['options'].append(opt)
								else:
									raise Exception(_('Invalid option-node found in "%s".') % filename)
					
					# create new group
					if len(group['options']):
						self.add_options_group(group['name'], group['info'])
						for o in group['options']:
							self.add_option(o)
					# add group to list
					#groups.append(group)
	
	def show_options_dialog (self):
		"""Show the OptionsDialog to edit this class' options."""
		dialog = OptionsDialog(490, 450)
		dialog.show_options_for_object(self)
		resp = se.run()
		if resp == gtk.RESPONSE_REJECT:	# TODO!
			dialog.reset_to_defaults()
		se.destroy()

# -----------------------------------------------------------------------
# OptionsDialog and UI-classes
# -----------------------------------------------------------------------

class ListOptionDialog (gtk.Dialog):
	"""An editing dialog used for editing options of the ListOption-type."""
	
	model = None
	tree = None
	buttonbox = None
	
	# call gtk.Dialog.__init__
	def __init__ (self):
		super(ListOptionDialog, self).__init__("Edit List", 
			flags=gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_NO_SEPARATOR,
			buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
				gtk.STOCK_OK, gtk.RESPONSE_OK))
		# set size
		self.resize(300, 370)
		self.set_keep_above(True)	# to avoid confusion
		# init vars
		self.model = gtk.ListStore(str)
		# create UI
		self.create_ui()
	
	def create_ui (self):
		"""Create the user-interface for this dialog."""
		# create outer hbox (tree|buttons)
		hbox = gtk.HBox()
		hbox.set_border_width(10)
		hbox.set_spacing(10)
		# create tree
		self.tree = gtk.TreeView(model=self.model)
		self.tree.set_headers_visible(False)
		self.tree.set_reorderable(True)
		#self.tree.set_grid_lines(gtk.TREE_VIEW_GRID_LINES_HORIZONTAL)
		col = gtk.TreeViewColumn('')
		cell = gtk.CellRendererText()
		#cell.set_property('cell-background', 'cyan')
		cell.set_property('foreground', 'black')
		col.pack_start(cell, False)
		col.set_attributes(cell, text=0)
		self.tree.append_column(col)
		self.tree.show()
		hbox.pack_start(self.tree, True, True)
		#sep = gtk.VSeparator()
		#sep.show()
		#hbox.add(sep)
		# create  buttons
		self.buttonbox = bb = gtk.VButtonBox()
		self.buttonbox.set_layout(gtk.BUTTONBOX_START)
		b1 = gtk.Button(stock=gtk.STOCK_ADD)
		b2 = gtk.Button(stock=gtk.STOCK_EDIT)
		b3 = gtk.Button(stock=gtk.STOCK_REMOVE)
		b1.connect('clicked', self.button_callback, 'add')
		b2.connect('clicked', self.button_callback, 'edit')
		b3.connect('clicked', self.button_callback, 'remove')
		bb.add(b1)
		bb.add(b2)
		bb.add(b3)
		self.buttonbox.show_all()
		#hbox.add(self.buttonbox)
		hbox.pack_end(self.buttonbox, False)
		# add everything to outer hbox and show it
		hbox.show()
		self.vbox.add(hbox)
		
	def set_list (self, lst):
		"""Set the list to be edited in this editor."""
		for el in lst:
			self.model.append([el])
	
	def get_list (self):
		"""Return the list that is currently being edited in this editor."""
		lst = []
		for i in self.model:
			lst.append(i[0])
		return lst
	
	def remove_selected_item (self):
		"""Remove the currently selected item."""
		sel = self.tree.get_selection()
		if sel:
			it = sel.get_selected()[1]
			if it:
				self.model.remove(it)
	
	def entry_dialog (self, default = ''):
		"""Show entry-dialog and return string."""
		entry = gtk.Entry()
		entry.set_text(default)
		entry.show()
		dlg = gtk.Dialog("Add/Edit Item", flags=gtk.DIALOG_DESTROY_WITH_PARENT,
			buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, 
			gtk.RESPONSE_OK))
		dlg.set_keep_above(True)
		dlg.vbox.add(entry)
		resp = dlg.run()
		ret = None
		if resp == gtk.RESPONSE_OK:
			ret = entry.get_text()
		dlg.destroy()
		return ret
				
	def button_callback (self, widget, id):
		if id == 'remove':
			self.remove_selected_item()
		if id == 'add':
			new = self.entry_dialog()
			if new != None:
				self.model.append([new])
		if id == 'edit':
			sel = self.tree.get_selection()
			if sel:
				it = sel.get_selected()[1]
				if it:
					new = self.entry_dialog(self.model.get_value(it, 0))
					if new != None:
						#self.model.append([new])
						self.model.set_value(it, 0, new)

class OptionsDialog (gtk.Dialog):
	"""A dynamic dialog for editing Options."""
	
	__shown_object = None
	
	def __init__ (self, width=500, height=500):
		# call gtk.Dialog.__init__
		super(OptionsDialog, self).__init__(_("Edit Options"),
			flags=gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_NO_SEPARATOR,
			buttons = (gtk.STOCK_CLOSE, gtk.RESPONSE_OK))
		# set size
		self.resize(width, height)
		self.set_keep_above(True)	# to avoid confusion
		self.set_border_width(10)
		# create attribs
		self.page_about		= None
		self.page_options	= None
		self.page_themes	= None
		self.vbox_editor	= None
		self.hbox_about		= None
		self.infotext		= None
		self.infoicon		= None
		# create theme-list
		self.liststore	= gtk.ListStore(object)
		self.tree		= gtk.TreeView(model=self.liststore)
		# create/add outer notebook
		self.main_notebook = gtk.Notebook()
		self.main_notebook.show()
		self.vbox.add(self.main_notebook)
		# create/init notebook pages
		self.create_about_page()
		self.create_themes_page()
		self.create_options_page()
		# crete tooltips-object
		self.tooltips = gtk.Tooltips()
	
	# "public" functions
	
	def reset_to_defaults (self):
		"""Reset all entries for the currently shown object to their default 
		values (the values the object has when it is first created).
		NOTE: This function resets ALL options, so BE CARFEUL!"""
		if self.__shown_object:
			for o in self.__shown_object.__options__:
				# set default value
				setattr(self.__shown_object, o.name, o.default)

	def set_info (self, name, info, copyright='', version='', icon=None):
		"""Update the "About"-page with the given information."""
		# convert infotext (remove EOLs and TABs)
		info = info.replace("\n", "")
		info = info.replace("\t", " ")
		# create markup
		markup = '\n<b><span size="xx-large">' + name + '</span></b>'
		if version:
			markup += '  <span size="large"><b>' + version + '</b></span>'
		markup += '\n\n'+info+'\n<span size="small">\n'+copyright+'</span>'
		self.infotext.set_markup(markup)
		# icon?
		if icon:
			# remove old icon
			if self.infoicon:
				self.infoicon.destroy()
			# set new icon
			self.infoicon = icon
			self.infoicon.set_alignment(0.0, 0.10)
			self.infoicon.show()
			self.hbox_about.pack_start(self.infoicon, 0, 1, 10)
		else:
			self.infoicon.hide()

	def show_options_for_object (self, obj):
		"""Update the OptionsEditor to show the options for the given Object.
		The Object needs to be an EditableOptions-subclass.
		NOTE: This needs heavy improvement and should use OptionGroups once
		      they exist"""
		self.__shown_object = obj
		# create notebook for groups
		notebook = gtk.Notebook()
		self.vbox_editor.add(notebook)
		for group in obj.__options_groups_ordered__:
			group_data = obj.__options_groups__[group]
			# create box for tab-page
			page = gtk.VBox()
			page.set_border_width(10)
			if group_data['info'] != '':
				info = gtk.Label(group_data['info'])
				info.show()
				info.set_alignment(0, 0)
				page.pack_start(info, 0, 0, 7)
				sep = gtk.HSeparator()
				sep.show()
				#page.pack_start(sep, 0, 0, 5)
			# create VBox for inputs
			box = gtk.VBox()
			box.show()
			box.set_border_width(5)
			# add box to page
			page.add(box)
			page.show()
			# add new notebook-page
			label = gtk.Label(group_data['label'])
			label.show()
			notebook.append_page(page, label)
			# and create inputs
			for option in group_data['options']:
				if option.hidden == False:
					val = getattr(obj, option.name) #obj.__dict__[option.name]
					w = self.get_widget_for_option(option, val)
					if w:
						box.pack_start(w, 0, 0)
						w.show()
		notebook.show()	
		# show/hide themes tab, depending on whether the screenlet uses themes
		if issubclass(obj.__class__, screenlets.Screenlet) and obj.uses_theme and obj.theme_name != '':
			self.show_themes_for_screenlet(obj)
		else:
			self.page_themes.hide()
	
	def show_themes_for_screenlet (self, obj):
		"""Update the Themes-page to display the available themes for the
		given Screenlet-object."""
		# list with found themes
		found_themes = []
		# now check all paths for themes
		for path in screenlets.SCREENLETS_PATH:
			p = path + '/' + obj.get_short_name() + '/themes'
			log.debug("Options path is %s" % p)
			#p = '/usr/local/share/screenlets/Clock/themes'	# TEMP!!!
			try:
				dircontent = os.listdir(p)
			except:
				log.warning(_("Path %s not found.") % p)
				continue
			# check all themes in path
			for name in dircontent:
				# load themes with the same name only once
				if found_themes.count(name):
					continue
				found_themes.append(name)
				# build full path of theme.conf
				theme_conf	= p + '/' + name + '/theme.conf'
				# if dir contains a theme.conf
				if os.access(theme_conf, os.F_OK):
					# load it and create new list entry
					ini = screenlets.utils.IniReader()
					if ini.load_conf(theme_conf):
						# check for section
						if ini.has_section('Theme'):
							# get metainfo from theme
							th_fullname	= ini.get_option('name', 
								section='Theme')
							th_info		= ini.get_option('info', 
								section='Theme')
							th_version	= ini.get_option('version', 
								section='Theme')
							th_author	= ini.get_option('author', 
								section='Theme')
							# create array from metainfo and add it to liststore
							info = [name, th_fullname, th_info, th_author, 
								th_version]
							self.liststore.append([info])
				else:
					# no theme.conf in dir? just add theme-name
					self.liststore.append([[name, '-', '-', '-', '-']])
				# is it the active theme?
				if name == obj.theme_name:
					# select it in tree
					log.debug(_("active theme is: %s") % name)
					sel = self.tree.get_selection()
					if sel:
						it = self.liststore.get_iter_from_string(\
							str(len(self.liststore)-1))
						if it:
							sel.select_iter(it)

	# UI-creation
	
	def create_about_page (self):
		"""Create the "About"-tab."""
		self.page_about = gtk.HBox()
		# create about box
		self.hbox_about = gtk.HBox()
		self.hbox_about.show()
		self.page_about.add(self.hbox_about)
		# create icon
		self.infoicon = gtk.Image()
		self.infoicon.show()
		self.page_about.pack_start(self.infoicon, 0, 1, 10)
		# create infotext
		self.infotext = gtk.Label()
		self.infotext.use_markup = True
		self.infotext.set_line_wrap(True)
		self.infotext.set_alignment(0.0, 0.0)
		self.infotext.show()
		self.page_about.pack_start(self.infotext, 1, 1, 5)
		# add page
		self.page_about.show()
		self.main_notebook.append_page(self.page_about, gtk.Label(_('About ')))	
	
	def create_options_page (self):
		"""Create the "Options"-tab."""
		self.page_options = gtk.HBox()
		# create vbox for options-editor
		self.vbox_editor	= gtk.VBox(spacing=3)
		self.vbox_editor.set_border_width(5)
		self.vbox_editor.show()
		self.page_options.add(self.vbox_editor)
		# show/add page
		self.page_options.show()
		self.main_notebook.append_page(self.page_options, gtk.Label(_('Options ')))

	def create_themes_page (self):
		"""Create the "Themes"-tab."""
		self.page_themes = gtk.VBox(spacing=5)
		self.page_themes.set_border_width(10)
		# create info-text list
		txt = gtk.Label(_('Themes allow you to easily switch the appearance of your Screenlets. On this page you find a list of all available themes for this Screenlet.'))
		txt.set_size_request(450, -1)
		txt.set_line_wrap(True)
		txt.set_alignment(0.0, 0.0)
		txt.show()
		self.page_themes.pack_start(txt, False, True)
		# create theme-selector list
		self.tree.set_headers_visible(False)
		self.tree.connect('cursor-changed', self.__tree_cursor_changed)
		self.tree.show()
		col = gtk.TreeViewColumn('')
		cell = gtk.CellRendererText()
		col.pack_start(cell, True)
		#cell.set_property('foreground', 'black')
		col.set_cell_data_func(cell, self.__render_cell)
		self.tree.append_column(col)
		# wrap tree in scrollwin
		sw = gtk.ScrolledWindow()
		sw.set_shadow_type(gtk.SHADOW_IN)
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		sw.add(self.tree)
		sw.show()
		# add vbox and add tree/buttons
		vbox = gtk.VBox()
		vbox.pack_start(sw, True, True)
		vbox.show()
		# show/add page
		self.page_themes.add(vbox)
		self.page_themes.show()
		self.main_notebook.append_page(self.page_themes, gtk.Label(_('Themes ')))
		
	def __render_cell(self, tvcolumn, cell, model, iter):
		"""Callback for rendering the cells in the theme-treeview."""
		# get attributes-list from Treemodel
		attrib = model.get_value(iter, 0)

		# set colors depending on state
		col = '555555'
		name_uc = attrib[0][0].upper() + attrib[0][1:]
		# create markup depending on info
		if attrib[1] == '-' and attrib[2] == '-':
			mu = '<b><span weight="ultrabold" size="large">' + name_uc + \
			'</span></b> (' + _('no info available') + ')'
		else:
			if attrib[1] == None : attrib[1] = '-'
			if attrib[2] == None : attrib[2] = '-'
			if attrib[3] == None : attrib[3] = '-'
			if attrib[4] == None : attrib[4] = '-'
			mu = '<b><span weight="ultrabold" size="large">' + name_uc + \
				'</span></b> v' + attrib[4] + '\n<small><span color="#555555' +\
				'">' + attrib[2].replace('\\n', '\n') + \
				'</span></small>\n<i><small>by '+str(attrib[3])+'</small></i>'
		# set markup
		cell.set_property('markup', mu)
	
	# UI-callbacks
	
	def __tree_cursor_changed (self, treeview):
		"""Callback for handling selection changes in the Themes-treeview."""
		sel = self.tree.get_selection()
		if sel:
			s = sel.get_selected()
			if s:
				it = s[1]
				if it:
					attribs = self.liststore.get_value(it, 0)
					if attribs and self.__shown_object:
						# set theme in Screenlet (if not already active)
						if self.__shown_object.theme_name != attribs[0]:
							self.__shown_object.theme_name = attribs[0]
	
	# option-widget creation (should be split in several classes)
	
	def get_widget_for_option (self, option, value=None):
		"""Return a gtk.*Widget with Label within a HBox for a given option."""
		hbox = gtk.HBox()
		label = gtk.Label()
		label.set_alignment(0.0, 0.0)
		label.set_label(option.label)
		label.set_size_request(180, 28)
		label.show()
		hbox.pack_start(label, 0, 1)
		widget = option.get_widget(value, self)
		if widget is not None:
			# the option is disabled then make it insensitive
			if not option.enabled:
				widget.set_sensitive(False)
				label.set_sensitive(False)
			#label.set_mnemonic_widget(widget)
			self.tooltips.set_tip(widget, option.desc)
			widget.show()
			# check if needs Apply-button
			if option.realtime == False:
				but = gtk.Button(_('Apply'), gtk.STOCK_APPLY)
				but.show()
				but.connect("clicked", self.apply_options_callback, 
					option, widget)
				b = gtk.HBox()
				b.show()
				b.pack_start(widget, 0, 0)
				b.pack_start(but, 0, 0)
				hbox.pack_start(b, 0, 0)
			else:
				#hbox.pack_start(widget, -1, 1)
				hbox.pack_start(widget, 0, 0)
		return hbox
	
	def read_option_from_widget (self, widget, option):
		"""Read an option's value from the widget and return it."""
		if widget.window is None:
			return False
		return option.get_value_from_widget(widget)
	
	# option-widget event-handling
	
	# TODO: custom callback/signal for each option?
	def options_callback (self, widget, optionobj):
		"""Callback for handling changed-events on entries."""
		log.debug(_("Changed: %s") % optionobj.name)
		if self.__shown_object:
			# if the option is not real-time updated,
			if optionobj.realtime == False:
				return False
			# read option
			val = self.read_option_from_widget(widget, optionobj)
			if val is not None:
				log.debug("SetOption: " + optionobj.name + "=" + str(val))
				
				# set option
				setattr(self.__shown_object, optionobj.name, val)
				
				# If a callback has been set and the callback wasn't called
				# from setattr because optionobj.always_trigger_callcack == False
				# then call the callback now. Please see Option.__init__ for more info.
				if optionobj.callback is not None and not optionobj.always_trigger_callback:
					optionobj.callback()
				
		return False

	def apply_options_callback (self, widget, optionobj, entry):
		"""Callback for handling Apply-button presses."""
		if self.__shown_object:
			# read option
			val = self.read_option_from_widget(entry, optionobj)
			if val is not None:
				# set option
				setattr(self.__shown_object, optionobj.name, val)
				
				# If a callback has been set and the callback wasn't called
				# from setattr because optionobj.always_trigger_callcack == False
				# then call the callback now. Please see Option.__init__ for more info.
				if optionobj.callback is not None and not optionobj.always_trigger_callback:
					optionobj.callback()
		return False

# create a dictionary of all option classes
OPTIONS_MAPPING = {"AccountOption": AccountOption,
	"BoolOption": BoolOption,
	"ColorOption": ColorOption,
	"DirectoryOption": DirectoryOption,
	"FileOption": FileOption,
	"FloatOption": FloatOption,
	"FontOption": FontOption,
	"ImageOption": ImageOption,
	"IntOption": IntOption,
	"ListOption": ListOption,
	"Option": Option,
	"StringOption": StringOption,
	"TimeOption": TimeOption}
