# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

#  Screenlets Class - (c) 2008 by Natan Yellin (Aantn) <aantny@gmail.com>
#  and Whise (Helder Fraga) <helder.fraga@hotmail.com>
#
#  Original author RYX (aka Rico Pfaus) <ryx@ryxperience.com> 

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

import glob

import dbus
import gtk
import pango

import screenlets
import screenlets.menu
from screenlets.options import *
from . import logger
from . import utils
from . import services
from .applet import AppletView
from .drawing import Drawing
from .menu import DefaultMenuItem
from .theme import Theme
from .notify import Notify
from .tooltip import Tooltip

gettext.textdomain('screenlets')
gettext.bindtextdomain('screenlets', screenlets.INSTALL_PREFIX +  '/share/locale')

def _(s):
	return gettext.gettext(s)

log = logger.get_default_logger()

class Screenlet (EditableOptions, Drawing):
	"""A Screenlet is a (i.e. contains a) shaped gtk-window that is
	fully invisible by default. Subclasses of Screenlet can render 
	their owner-drawn graphics on fully transparent background."""
	
	# default meta-info for Screenlets
	__name__	= _('No name set for this Screenlet')
	__version__	= '0.0'
	__author__	= _('No author defined for this Screenlet')
	__desc__	= _('No info set for this Screenlet')
	__requires__	= []		# still unused
	#__target_version__ = '0.0.0'
	#__backend_version__ = '0.0.1'
	
	# legacy support 
	scale = 1.0
	
	has_started	= False
	
	def __init__ (self, id='', width=100, height=100, uses_theme=True,
		path=os.getcwd(), drag_drop=False, session=None, enable_saving=True,
		service_class=services.ScreenletService, uses_pango=False,
		uses_shape=True, is_sizeable=True, resize_on_scroll=True,
		ask_on_option_override=True, server_blacklist=[], **keyword_args):
		"""Constructor - should only be subclassed"""
		
		# call the EditableOptions and AppletView constructors
		super(Screenlet, self).__init__(session.backend, id, enable_saving)
				
		# initialize attributes
		# TODO: document all attributes
		self.id				= id
		self.drag_drop		= drag_drop
		self.service_class	= service_class
		self.is_dragged		= False	# True if the screenlet is being dragged
		self.dragging_over	= False	# True if something is dragged over
		self.mouse_is_over	= False
		self.has_focus		= False
		self.mousex			= 0
		self.mousey			= 0
		self.tooltip		= None	# Used by show_tooltip()
		self.notify			= None	# Used by show_notification()
		self.server_blacklist = server_blacklist
		
		# find the theme path
		theme_path = []
		for p in screenlets.SCREENLETS_PATH:
			theme_path.append(p+"/"+self.get_short_name()+"/themes/")
			
		# attributes
		self.theme = None							# the current theme object
		self.theme_path = theme_path				# paths to search for themes
		self.uses_theme = uses_theme				# flag indicating whether theming is enabled
		self.disable_updates = False				# to temporarily avoid refresh/reshape
		self.allow_option_override 	= True			# if False, overrides are ignored
		self.ask_on_option_override	= True			# if True, overrides need confirmation
		
		# window shaping attributes
		self.__need_to_update_shape = False
		self.__shape_bitmap_width = 0
		self.__shape_bitmap_height = 0
		
		# set other attributes without calling __setattr__
		self.__dict__['theme_name'] = ""			# name of currently active theme
		self.__dict__['uses_shape'] = uses_shape	# true if this screenlet has a custom shape
	
		# todo: load from server
		self.skip_taskbar = True
		self.skip_pager = True
		
		# init properties
		self.disable_updates  = True
		self.disable_server_menu_callbacks = False
		self.uses_shape		  = uses_shape
		self.resize_on_scroll = resize_on_scroll
		self.is_dragged 	  = False
		self.session 		  = session
		self.service		  = None
		self.__path__ 		  = path
		self.__dict__["width"]	= width
		self.__dict__["height"]	= height
		self.scale	= 1.0
		
		# Deprecated attributes
		# TODO: Set tell the server when they change and request a placement change?
		self.y		= 0
		self.x		= 0
		
		# "editable" options, first create a group
		self.add_options_group('Screenlet', 
			_('The basic settings for this Screenlet-instance.'))
		# if this Screenlet uses themes, add theme-specific options
		# (NOTE: this option became hidden with 0.0.9 and doesn't use
		# get_available_themes anymore for showing the choices)
		if self.uses_theme:
			self.add_option(StringOption('Screenlet', 'theme_name', 
				'default', '', '', hidden=True))
		# create/add options
		self.add_option(BoolOption('Screenlet', 'allow_option_override', 
			self.allow_option_override, _('Allow overriding Options'), 
			_('Allow themes to override options in this screenlet ...')))
		self.add_option(BoolOption('Screenlet', 'ask_on_option_override', 
			self.ask_on_option_override, _('Ask on Override'), 
			_('Show a confirmation-dialog when a theme wants to override ')+\
			_('the current options of this Screenlet ...')))
		
		# create a gtk.Layout and connect signals
		self.layout = gtk.Layout()
		
		self.layout.set_size(width, height)
		self.layout.set_app_paintable(True)
		screenlets.alpha_screen_changed(self.layout)
		
		self.layout.set_events(gtk.gdk.ALL_EVENTS_MASK)
		self.layout.set_flags(gtk.CAN_FOCUS)
		self.layout.grab_focus()
		
		self.layout.connect("button-press-event",	self.button_press)
		self.layout.connect("button-release-event",	self.button_release)
		self.layout.connect("composited-changed",	self.composite_changed)
		self.layout.connect("delete_event",			self.delete_event)
		self.layout.connect("destroy",				self.destroy)
		self.layout.connect("expose-event",			self.expose)
		self.layout.connect("focus-in-event",		self.focus_in_event)
		self.layout.connect("focus-out-event",		self.focus_out_event)
		self.layout.connect("enter-notify-event",	self.enter_notify_event)
		self.layout.connect("key-press-event",		self.key_press) # TODO: use keyword-attrib to activate?
		self.layout.connect("leave-notify-event",	self.leave_notify_event)
		self.layout.connect("motion-notify-event",	self.motion_notify_event)
		self.layout.connect("realize",				self.realize_event)
		self.layout.connect("screen-changed",		screenlets.alpha_screen_changed)
		self.layout.connect("scroll-event",			self.scroll_event)
		self.layout.connect("unmap-event",			self.hide_event)
		
		# create pango layout, if active
		if uses_pango:
			self.p_context = self.layout.get_pango_context()
			if self.p_context:
				self.p_layout = pango.Layout(self.p_context)
				self.p_layout.set_font_description(pango.FontDescription("Sans 12"))
		
		self.layout.show_all()
		
		self.main_view = AppletView(self.id, self.width, self.height, screenlets.COMMAND_LINE_OPTIONS.SERVER_NAME)
		self.main_view.add(self.layout)
		
		# Legacy support
		self.window = self.content = self.layout
		
		# create menu
		self.menu = gtk.Menu()

	def __setattr__ (self, name, value):
		# Note: Many of the on_* functions are called from other functions
		# when attributes change (and not from here)
		# For example, on_scale is called from the handler for expose events
		# TODO: Should the on_* functions be handled in a more orderly fashion?
		self.on_before_set_atribute(name, value)
		
		if name == "theme_name" and self.uses_theme:
			log.debug(_("Loading new theme: ") + value)
			path = self.find_theme(value)
			if path:
				log.debug(_("Found: ") + str(path))
				self.load_theme(path)
				self.redraw_canvas()
				self.update_shape()
		elif name == "uses_shape":
			if value == True:
				self.update_shape()
			# TODO: Reset/Clear shape if value == False
		elif name == "width" or name == "height":
			# TODO: Add a set_size method to AppletView
			self.main_view.set_size(self.width, self.height)
			
		super(Screenlet, self).__setattr__(name, value)
		self.on_after_set_atribute(name, value)
	
	#-----------------------------------------------------------------------
	# Screenlet's public functions
	#-----------------------------------------------------------------------
	
	def add_default_menuitems (self, flags=DefaultMenuItem.STANDARD):
		"""Appends the default menu-items to self.menu. You can add on OR'ed
		   flag with DefaultMenuItems you want to add."""		
		menu = self.menu
		
		# children already exist? add separator
		if len(menu.get_children()) > 0:
			self.add_menuitem("", "-")
		# EXPERIMENTAL:
		if flags & DefaultMenuItem.XML:
			# create XML-menu from screenletpath/menu.xml
			xfile = self.get_screenlet_dir() + "/menu.xml"
			xmlmenu = screenlets.menu.create_menu_from_file(xfile, 
				self.menuitem_callback)
			if xmlmenu:
				self.menu = xmlmenu

		# create theme-selection menu
		if flags & DefaultMenuItem.THEMES:
			themes_item = screenlets.menu.ImageMenuItem(gtk.STOCK_SELECT_COLOR,
				_("Theme"))
			themes_item.show()
			themes_menu = gtk.Menu()
			menu.append(themes_item)
			themes_item.set_submenu(themes_menu)
			# create theme-list from theme-directory
			lst = self.get_available_themes()
			for tname in lst:
				item = gtk.MenuItem(tname)
				item.connect("activate", self.menuitem_callback, "theme:"+tname)
				item.show()
				themes_menu.append(item)
			screenlets.menu.add_menuitem(menu, "-", self.menuitem_callback, "")
		
		# add server-name item
		if flags & DefaultMenuItem.SERVER_NAME:
			# TODO: Use a better image for this.
			server_item = screenlets.menu.ImageMenuItem(gtk.STOCK_EXECUTE,
				_("Move to"))
			server_item.show()
			menu.append(server_item)
			
			self.server_menu = gtk.Menu()
			server_item.set_submenu(self.server_menu)
			
			bus = dbus.SessionBus()
			
			group = None # Create a new group for the first radio item.
			for server in self.main_view.get_running_servers():
				if server in self.server_blacklist:
					continue
				
				proxy_obj = bus.get_object("org.UniversalApplets." + server,
					"/org/UniversalApplets/" + server, "org.UniversalApplets." + server)
				text = proxy_obj.get_server_location_description()
				item = gtk.RadioMenuItem(group, text)
				if group is None:
					group = item # Set the group to the first radio item.
				item.show()
				
				# check if the server is the currently active server
				if server == screenlets.COMMAND_LINE_OPTIONS.SERVER_NAME:
					item.set_active(True)
					self.server_menu.insert(item, 0)
				else:
					self.server_menu.append(item)
				
				item.connect("activate", self.menuitem_callback, "server:" + server)
		
		# add Settings item
		if flags & DefaultMenuItem.PROPERTIES:
			screenlets.menu.add_image_menuitem(menu, gtk.STOCK_PREFERENCES,
				_("Preferences"), self.menuitem_callback, "options")
		
		# add info item
		if flags & DefaultMenuItem.INFO:
			screenlets.menu.add_menuitem(menu, "-", self.menuitem_callback, "")
			screenlets.menu.add_image_menuitem(menu, gtk.STOCK_DIALOG_INFO,
				_("Info"), self.menuitem_callback, "info")
		
		# add delete item; TODO: Either rename "delete" to "remove" or change the icon 
		if flags & DefaultMenuItem.DELETE:
			screenlets.menu.add_menuitem(menu, "-", self.menuitem_callback, "")
			screenlets.menu.add_image_menuitem(menu, gtk.STOCK_REMOVE,
				_("Delete Screenlet"), self.menuitem_callback, "delete")
		
		# add Quit item
		if flags & DefaultMenuItem.QUIT:
			screenlets.menu.add_menuitem(menu, "-", self.menuitem_callback, "")
			screenlets.menu.add_image_menuitem(menu, gtk.STOCK_CLOSE,
				_("Quit this %s") % self.get_short_name(), self.menuitem_callback,
				"quit_instance")
		
		# add Quit-all item
		if flags & DefaultMenuItem.QUIT_ALL:
			screenlets.menu.add_menuitem(menu, "-", self.menuitem_callback, "")
			screenlets.menu.add_image_menuitem(menu, gtk.STOCK_QUIT, _("Quit all %ss")
				% self.get_short_name(), self.menuitem_callback, "quit")

	def add_menuitem (self, id, label, callback=None):
		"""Simple way to add menuitems to a right-click menu.
		This function wraps screenlets.menu.add_menuitem.
		For backwards compatibility, the order of the parameters
		to this function is switched."""
		if callback is None:
			callback = self.menuitem_callback
		# call menu.add_menuitem
		return screenlets.menu.add_menuitem(self.menu, label, callback, id)
	
	def add_submenuitem (self, id, label, lst, callback=None):
		"""Simple way to add submenuitems to the right-click menu through a list."""
		submenu = gtk.MenuItem(label)
		submenu.show()
		sub_menu = gtk.Menu()
		self.menu.append(submenu)
		submenu.set_submenu(sub_menu)
		
		for tname in lst:
			item = gtk.MenuItem(tname)
			item.connect("activate", self.menuitem_callback, 
				tname)
			item.show()
			sub_menu.append(item)

		return submenu
	
	def create_drag_icon (self):
		"""Create drag-icon and -mask for drag-operation. Returns a 2-tuple
		with the icon and the mask. To supply your own icon you can use the
		on_create_drag_icon-handler and return the icon/mask as 2-tuple."""
		w = self.width
		h = self.height
		icon, mask = self.on_create_drag_icon()
		if icon is None:
			# create icon
			icon = gtk.gdk.Pixmap(self.content.bin_window, w, h)
			ctx = icon.cairo_create()
			clear_cairo_context(ctx)
			self.on_draw(ctx)
		if mask is None:
			# create mask
			mask = gtk.gdk.Pixmap(self.content.bin_window, w, h)
			ctx = mask.cairo_create()
			clear_cairo_context(ctx)
			self.on_draw_shape(ctx)
		return (icon, mask)
	
	def find_theme (self, name):
		"""Find the first occurence of a theme and return its global path."""
		for p in self.theme_path:
			fpath = p + name
			if os.path.isdir(fpath):
				return fpath
		return None
	
	def finish_loading(self):
		"""Called to finish the screenlet's initialization.
		This is called by ScreenletSession after the screenlet's
		__init__ method is called."""
		# activate applet mode
		self.main_view.set_active(True)
		
		# drag/drop support (NOTE: still experimental and incomplete)
		if self.drag_drop:
			self.main_view.plug.drag_dest_set(gtk.DEST_DEFAULT_MOTION |
				gtk.DEST_DEFAULT_DROP, #gtk.DEST_DEFAULT_ALL, 
				[("text/plain", 0, 0), 
				("image", 0, 1),
				("text/uri-list", 0, 2)], 
				gtk.gdk.ACTION_COPY)
			self.main_view.plug.connect("drag_data_received", self.drag_data_received)
			self.main_view.plug.connect("drag-begin", self.drag_begin)
			self.main_view.plug.connect("drag-end", self.drag_end)
			self.main_view.plug.connect("drag-motion", self.drag_motion)
			self.main_view.plug.connect("drag-leave", self.drag_leave)
		
		self.disable_updates = False
		# if we have an id and a service-class, register our service
		if self.id and self.service_class:
			self.register_service(self.service_class)
			# notify service about adding this instance
			self.service.instance_added(self.id)
		self.has_started = True	
		self.on_init()
		
		# update shape
		self.update_shape()
		
	def get_short_name (self):
		"""Return the short name of this screenlet. This returns the classname
		of the screenlet without trailing "Screenlet". Please always use
		this function if you want to retrieve the short name of a Screenlet."""
		return self.__class__.__name__[:-9]
		
	def get_screenlet_dir (self):
		"""Return the name of this screenlet's personal directory."""
		p = utils.find_first_screenlet_path(self.get_short_name())
		if p:
			return p
		else:
			if self.__path__ != '':
				return self.__path__
			else:
				return os.getcwd()
	
	def get_available_themes (self):
		"""Returns a list with the names of all available themes in this
			Screenlet's theme-directory."""
		lst = []
		for p in self.theme_path:
			if os.path.isdir(p):
				#dirname = self.get_theme_dir()
				dirlst = glob.glob(p + '*')
				dirlst.sort()
				tdlen = len(p)
				for fname in dirlst:
					dname = fname[tdlen:]
					# TODO: check if it's a dir
					lst.append(dname)
		return lst
	
	def get_theme_dir (self):
		"""Return the name of this screenlet's personal theme-dir.
		(Only returns the dir under the screenlet's location"""
		#return self.get_screenlet_dir() + "/themes/"
		return self.theme_path
	
	def load_theme (self, path):
		"""Load a theme for this Screenlet from the given path. NOTE: 
		load_theme does NOT call redraw_canvas and update_shape!!!!! To do all 
		in one call, set the attribute self.theme_name instead."""
		if self.theme:
			self.theme.free()
		self.theme = Theme(path)
		# check for errors
		if self.theme.loaded == False:
			log.error(_("Could not load theme: ") + path)
			# todo: free theme?
			self.theme = None
		else:
			# call user-defined handler
			# if override options is allowed, apply them
			if self.allow_option_override:
				if self.theme.has_overrides():
					if not self.ask_on_option_override or screenlets.show_question(self,
						_('This theme wants to override your settings for this Screenlet. Do you want to allow that?')):
						self.apply_option_overrides(self.theme.get_option_overrides())
			
			self.on_load_theme()
	
	def register_service (self, service_classobj):
		"""Register or create the given ScreenletService-(sub)class as the new 
		service for this Screenlet. If self is not the first instance in the
		current session, the service from the first instance will be used 
		instead and no new service is created."""
		if self.session:
			if len(self.session.instances) == 0:
				# if it is the basic service, add name to call
				if service_classobj == services.ScreenletService:#BUG
					self.service = service_classobj(self, self.get_short_name())
				else:
					# else only pass this screenlet
					self.service = service_classobj(self)
			else:
				self.service = self.session.instances[0].service
			# TODO: throw exception??
			return True
		return False
	
	def show (self):
		"""Show this Screenlet's underlying gtk.Layout"""
		self.content.show()
		self.on_show()
	
	def show_options_dialog (self):
		"""Show the EditableSettingsDialog for this Screenlet."""
		se = OptionsDialog(490, 450)
		img = gtk.Image()
		try:
			d = self.get_screenlet_dir()
			if os.path.isfile(d + '/icon.svg'):
				icn = gtk.gdk.pixbuf_new_from_file(d + '/icon.svg')
			elif os.path.isfile(d + '/icon.png'):
				icn = gtk.gdk.pixbuf_new_from_file(d + '/icon.png')
			img.set_from_pixbuf(icn)
		except:
			img.set_from_stock(gtk.STOCK_PREFERENCES, 5)
		se.set_title(self.__name__)
		se.set_info(self.__name__, self.__desc__, '(c) ' + self.__author__, 
			version='v' + self.__version__, icon=img)
		se.show_options_for_object(self)
		resp = se.run()
		if resp == gtk.RESPONSE_REJECT:	# TODO!!!!!
			se.reset_to_defaults()
		else:
			self.redraw_canvas()
			self.update_shape()
		se.destroy()
	
	def redraw_canvas (self, widget=None):
		"""Redraw the entire Screenlet's window area."""			
		# if updates are disabled, just exit
		if self.disable_updates:
			return
			
		if widget is None:
			widget = self.layout
		
		x, y, w, h = widget.get_allocation()
		rect = gtk.gdk.Rectangle(x, y, w, h)
		try:
			if widget.window:
				widget.window.invalidate_rect(rect, True)
				widget.window.process_updates(True)
		except AttributeError, ex:
			if widget.bin_window:
				log.info("Drawing to bin window")
				widget.bin_window.invalidate_rect(rect, True)
				widget.bin_window.process_updates(True)	
	
	def redraw_canvas_area (self, x, y, width, height):
		"""Redraw the given Rectangle (x, y, width, height) within the 
		current Screenlet's window."""
		# if updates are disabled, just exit
		if self.disable_updates:
			return
		widget = self.layout
		if widget:
			rect = gtk.gdk.Rectangle(x, y, width, height)
			if widget.window is not None:
				widget.window.invalidate_rect(rect, True)
				widget.window.process_updates(True)
	
	def redraw_canvas_and_update_shape (self):
		self.redraw_canvas()
		self.update_shape()
		
	def update_shape (self):
		"""Update window shape. Only call this when shape has changed
		because it is very resource intensive if ran too often."""
		if self.disable_updates or not self.uses_shape:
			return False
		
		self.on_update_shape()
		
		# w = self.plug.allocation.width
		# h = self.plug.allocation.height
		w = self.window.allocation.width
		h = self.window.allocation.height
		
		# if 0 return to avoid crashing
		if w==0 or h==0: return False
		# if size changed, recreate shape bitmap
		if w != self.__shape_bitmap_width or h != self.__shape_bitmap_height:
			self.__shape_bitmap = screenlets.create_empty_bitmap(w, h)
			self.__shape_bitmap_width = w
			self.__shape_bitmap_height = h
			
		# create context
		ctx = self.__shape_bitmap.cairo_create()
		self.clear_cairo_context(ctx)
		
		# shape the window acording if the window is composited  or not
		if self.window.is_composited():
			log.debug(_("Updating input shape"))
			self.on_draw_shape(ctx)
			self.main_view.set_shape(self.__shape_bitmap, True)
		else:
			try:
				self.on_draw_shape(ctx)
			except:
				self.on_draw(ctx)
			log.debug(_("Updating window shape"))
			self.main_view.set_shape(self.__shape_bitmap, False)
	
	def show_notification (self, text):
		"""Show notification window at current mouse position."""
		if self.notify is None:
			self.notify = Notify()
			self.notify.text = text
			self.notify.show()

	def hide_notification (self):
		"""hide notification window"""
		if self.notify is not None:
			self.notify.hide()
			self.notify = None

	def show_tooltip (self, text, tooltipx, tooltipy):
		"""Show tooltip window at current mouse position."""
		if self.tooltip is None:
			self.tooltip = Tooltip(300, 400)
			self.tooltip.text = text
			self.tooltip.x    = tooltipx
			self.tooltip.y    = tooltipy
		self.tooltip.show()

	def hide_tooltip (self):
		"""hide tooltip window"""
		if self.tooltip is not None:
			self.tooltip.hide()
			self.tooltip = None
		
	# ----------------------------------------------------------------------
	# Screenlet's event-handler dummies
	# ----------------------------------------------------------------------
	
	def on_delete (self):
		"""Called when the Screenlet gets deleted. Return True to cancel.
		TODO: sometimes not properly called"""
		return not screenlets.show_question(self, _("To quit all %s's, use 'Quit' instead. ") % self.__class__.__name__ +\
			_('Really delete this %s and its settings?') % self.get_short_name())
		"""return not show_question(self, 'Deleting this instance of the '+\
				self.__name__ + ' will also delete all your personal '+\
				'changes you made to it!! If you just want to close the '+\
				'application, use "Quit" instead. Are you sure you want to '+\
				'delete this instance?')
		return False"""
	
	# TODO: on_drag
	# TODO: on_drag_end
	
	def on_after_set_atribute(self,name, value):
		"""Called after setting screenlet atributes"""
		pass
	
	def on_before_set_atribute(self,name, value):
		"""Called before setting screenlet atributes"""
		pass
	
	def on_composite_changed(self):
		"""Called when composite state has changed"""
		pass
	
	def on_create_drag_icon (self):
		"""Called when the screenlet's drag-icon is created. You can supply
		your own icon and mask by returning them as a 2-tuple."""
		return (None, None)

	def on_drag_begin (self, drag_context):
		"""Called when the Screenlet gets dragged."""
		pass
	
	def on_drag_enter (self, drag_context, x, y, timestamp):
		"""Called when something gets dragged into the Screenlets area."""
		pass
	
	def on_drag_leave (self, drag_context, timestamp):
		"""Called when something gets dragged out of the Screenlets area."""
		pass
	
	def on_draw (self, ctx):
		"""Callback for drawing the Screenlet's window - override
		in subclasses to implement your own drawing."""
		pass
	
	def on_draw_shape (self, ctx):
		"""Callback for drawing the Screenlet's shape - override
		in subclasses to draw the window's input-shape-mask."""
		pass
	
	def on_drop (self, x, y, sel_data, timestamp):
		"""Called when a selection is dropped on this Screenlet."""
		return False
		
	def on_focus (self, event):
		"""Called when the Screenlet's window receives focus."""
		pass
	
	def on_hide (self):
		"""Called when the Screenlet gets hidden."""
		pass
	
	def on_init (self):
		"""@DEPRECATED.
		Options are now loaded when they're added.
		You may do all of your initialization in __init__."""
		pass
	
	def on_key_down (self, keycode, keyvalue, event=None):
		"""Called when a key is pressed within the screenlet's window."""
		pass
	
	def on_load_theme (self):
		"""Called when the theme is reloaded (after loading, before redraw)."""
		pass
	
	def on_menuitem_select (self, id):
		"""Called when a menuitem is selected."""
		pass
	
	def on_mouse_down (self, event):
		"""Called when a buttonpress-event occured in Screenlet's window. 
		Returning True causes the event to be not further propagated."""
		return False
	
	def on_mouse_enter (self, event):
		"""Called when the mouse enters the Screenlet's window."""
		pass
		
	def on_mouse_leave (self, event):
		"""Called when the mouse leaves the Screenlet's window."""
		pass

	def on_mouse_move(self, event):
		"""Called when the mouse moves in the Screenlet's window."""
		pass
	
	def on_mouse_up (self, event):
		"""Called when a buttonrelease-event occured in Screenlet's window. 
		Returning True causes the event to be not further propagated."""
		return False
	
	def on_quit (self):
		"""Callback for handling destroy-event. Perform your cleanup here!"""
		return True
	
	def on_realize (self):
		""""Callback for handling the realize-event."""
	
	def on_scale (self):
		"""Called when Screenlet.scale is changed."""
		pass
	
	def on_scroll_up (self):
		"""Called when mousewheel is scrolled up (button4)."""
		pass

	def on_scroll_down (self):
		"""Called when mousewheel is scrolled down (button5)."""
		pass
	
	def on_show (self):
		"""Called when the Screenlet gets shown after being hidden."""
		pass
	
	def on_switch_widget_state (self, state):
		"""Called when the Screenlet enters/leaves "Widget"-state.
		This should be deprecated and removed."""
		pass
	
	def on_unfocus (self, event):
		"""Called when the Screenlet's window loses focus."""
		pass

	def on_update_shape(self):
		"""Called when the Screenlet's window is updating shape"""
		pass
		
	# ----------------------------------------------------------------------
	# Screenlet's event-handlers for GTK-events
	# ----------------------------------------------------------------------
	
	def button_press (self, widget, event):
		if self.has_focus == None:
			self.focus_in_event(self, None)
		
		if self.on_mouse_down(event):
			return True
		elif event.button == 3:
			self.menu.popup(None, None, None, event.button, event.time)
			return True
		else:
			return False
	
	def button_release (self, widget, event):
		if self.on_mouse_up(event):
			return True
		else:
			return False
	
	def delete_event (self, widget, event, data=None):
		# cancel event?
		if self.on_delete() == True:
			log.info(_("Canceled delete_event"))
			return True
		else:
			return False

	def destroy (self, widget, data=None):
		# call user-defined on_quit-handler
		self.on_quit()
		self.session.quit_instance(self.id)
	
	def drag_begin (self, widget, drag_context):
		log.debug(_("Start drag"))
		self.is_dragged = True
		self.on_drag_begin(drag_context)
		#return False
	
	def drag_data_received (self, widget, dc, x, y, sel_data, info, timestamp):
		return self.on_drop(x, y, sel_data, timestamp)
	
	def drag_end (self, widget, drag_context):
		log.debug(_("End drag"))
		self.is_dragged = False
		return False
	
	def drag_motion (self, widget, drag_context, x, y, timestamp):
		log.debug(_("Drag motion"))
		if self.dragging_over == False:
			self.dragging_over = True
			self.on_drag_enter(drag_context, x, y, timestamp)
		return False
	
	def drag_leave (self, widget, drag_context, timestamp):
		self.dragging_over = False
		self.on_drag_leave(drag_context, timestamp)
		return
	
	def enter_notify_event (self, widget, event):
		# check if the mouse really entered the screenlet's
		# area.
		if not self.mouse_is_over:
			self.mouse_is_over = True
			self.on_mouse_enter(event)
	
	def composite_changed (self, widget):
		"""this callback is called when the composite state changes"""
		self.update_shape()
		self.redraw_canvas()
		self.on_composite_changed()
	
	def expose (self, widget, event):
		# when the scale changes we'll get an expose event to redraw the applet
		# find the (new) scale
		old_scale = self.scale
		new_scale = float(widget.allocation.width)/self.width
		
		# check if the scale changed
		if old_scale != new_scale:
			# check if the server scaled the applet correctly
			if new_scale != float(widget.allocation.height)/self.height:
				log.warning("WARNING: The server improperly resized the applet.")
			
			# set the new scale and call the custom handler
			self.scale = new_scale
			self.on_scale()
			
			# now that the scale has changed, we need to update the shape.
			# the actual update will be done at the end of this function after we've
			# drawn onto the window. (the server uses our drawn image to update it's
			# shape. it doesn't (yet) use the shape that we set on the socket.)
			self.__need_to_update_shape = True
		
		# if updates are disabled then just return
		if self.disable_updates:
			return
			
		log.debug("Expose event")
		
		ctx = widget.bin_window.cairo_create()
		# clear context
		self.clear_cairo_context(ctx)
		# set a clip region for the expose event
		ctx.rectangle(event.area.x, event.area.y,
			event.area.width, event.area.height)
		ctx.clip()
		
		# scale context
		#ctx.scale(self.scale, self.scale)
		
		# call drawing method
		self.on_draw(ctx)
		
		# if we haven't set the window shape yet then do it now
		if self.__need_to_update_shape:
			self.update_shape()
			self.__need_to_update_shape = False
		return False
	
	def focus_in_event (self, widget, event):
		if self.skip_taskbar==False or self.skip_pager==False \
			or self.is_dragged==True or event is None:
			# Screenlet always gets focus after being dragged so this is a good method
			# to control the end of a move_drag operation
			self.is_dragged = False
			self.has_focus = True
			self.on_focus(event)

	def focus_out_event (self, widget, event):
		if self.is_dragged == False:
			self.has_focus = False
			self.on_unfocus(event)
	
	def hide_event (self, widget, event):
		self.on_hide()
		
	def key_press (self, widget, event):
		"""Handle keypress events, needed for in-place editing."""
		self.on_key_down(event.keyval, event.string, event)
	
	def leave_notify_event (self, widget, event):
		#self.__mouse_inside = False
		#self.is_dragged = False
		self.mouse_is_over = False
		self.on_mouse_leave(event)
	
	def menuitem_callback (self, widget, id):
		if id == "delete":
			if not self.on_delete():
				# remove instance
				self.session.delete_instance (self.id)
				# notify about being removed (does this get sent?)
				self.service.instance_removed(self.id)
		elif id[:6]=="theme:":
			log.debug(_("Screenlet: Set theme %s") % id[6:])
			# set theme
			self.theme_name = id[6:]
		elif id[:8] == "setting:":
			# set a boolean option to the opposite state
			try:
				if type(self.__dict__[id[8:]]) == bool:
					self.__dict__[id[8:]] = not self.__dict__[id[8:]]	# UNSAFE!!
			except:
				log.error(_("Cannot set missing or non-boolean value '")\
					+ id[8:] + "'")
		elif id[:7] == "server:" and not self.disable_server_menu_callbacks:
			# set self.disable_server_menu_callbacks to True so that gtk wont recursively
			# call this function when we (de)activate the old server's menuitem
			self.disable_server_menu_callbacks = True
			
			old_server = self.main_view.get_server()
			
			log.debug(_("Changing servers to ") + id[7:])
			self.main_view.set_server(id[7:])
			
			for item in self.server_menu.get_children():
				label = item.get_children()[0].get_label()
				if label == id[7:]:
					item.set_active(True)
				elif label == old_server:
					item.set_active(False)
					
			
			# reenable callbacks for the server menu
			self.disable_server_menu_callbacks = False
			
		elif id in ("info", "about", "settings", "options", "properties"):
			# show settings dialog
			self.show_options_dialog()
		elif id == "quit_instance":
			self.main_view.plug.destroy()
		elif id == "quit":
			self.session.quit()
		self.on_menuitem_select(id)
		return False

	def map_event(self, widget, event):
		self.on_map()

	def unmap_event(self, widget, event):
		self.on_unmap()

	def motion_notify_event(self, widget, event):
		self.__dict__['mousex'] = event.x / self.scale
		self.__dict__['mousey'] = event.y / self.scale
		
		self.on_mouse_move(event)
	
	def realize_event (self, widget):		
		"""called when window has been realized"""
		#self.update_shape()
		if widget.window:
			widget.window.set_back_pixmap(None, False)	# needed?
		
		self.on_realize()
	
	def scroll_event (self, widget, event):
		if event.direction == gtk.gdk.SCROLL_UP:
			if self.has_focus and self.is_sizable and self.resize_on_scroll: self.scale = self.scale +0.1
			self.on_scroll_up()
		elif event.direction == gtk.gdk.SCROLL_DOWN:
			if self.has_focus and self.is_sizable and self.resize_on_scroll: self.scale = self.scale -0.1
			self.on_scroll_down()
