# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Screenlets Session Module
# (c) 2008 Natan Yellin (Aantn) <aantny@gmail.com>
#
# Original author Rico Pfaus (RYX) <ryx@ryxperience.com>

# INFO:
# This module contains the ScreenletSession-class which handles the lower-level
# things like startup, multiple instances and sessions. It should also become
# the interface for load/save operations. The ScreenletSession is further
# responsible for handling command-line args to the Screenlet and should maybe
# offer some convenient way of setting Screenlet-options via commandline (so
# one can do "NotesScreenlet --theme_name=green --scale=0.5" and launch the
# Note with the given theme and scale)..
#
#
# INFO:
# - When a screenlet gets launched:
#   - the first instance of a screenlet creates the Session-object (within the
#     __main__-code)
#   - the session object investigates the config-dir for the given Screenlet
#     and restores available instances
#   - else (if no instance was found) it simply creates a new instance of the 
#     given screenlet and runs its mainloop
# - the --session argument allows setting the name of the session that will be
#   used by the Screenlet (to allow multiple configs for one Screenlet)
#
# TODO:
# - set attributes via commandline??
# - fix a possible error if the container and the screenlet change their settings at the same time
# - either fix the name param or get rid of it
# - quit the session if the last instance is closed

# please see http://www.python.org/dev/peps/pep-0328/#guido-s-decision
from __future__ import absolute_import

import os
from stat import S_IRWXU, S_IRWXG, S_IRWXO

import dbus
import gtk
import gettext

import screenlets
from . import backend
from . import logger
from . import services
from . import utils

gettext.textdomain('universal-applets')
gettext.bindtextdomain('universal-applets', screenlets.INSTALL_PREFIX +  '/share/locale')

_ = gettext.gettext

# set up logger
log = logger.get_default_logger()

class ScreenletSession (object):
	"""The ScreenletSession manages instances of a Screenlet and handles
	saving/restoring options. Each Screenlet contains a reference to its 
	session. Multiple instances of the same Screenlet share the same
	session-object."""
	
	def __init__ (self, screenlet_classobj, backend_type='caching', name='default'):
		"""Constructor"""
		super(ScreenletSession, self).__init__()
		# check type
		if not screenlet_classobj.__name__.endswith('Screenlet'):
			# TODO: also check for correct type (Screenlet-subclass)!!
			raise Exception(_("""ScreenletSession.__init__ has to be called with a
			 valid Screenlet-classobject as first argument!"""))
		
		# init properties
		self.name		= name
		self.screenlet	= screenlet_classobj
		self.instances	= []
		
		# set name of session (see SESSION_NAME in __init__.py for default value)
		self.name = screenlets.COMMAND_LINE_OPTIONS.SESSION_NAME
		log.debug("session.py : session name = : %s" %self.name)

		# set session path (and create dir-tree if not existent)
		p = screenlet_classobj.__name__[:-9] + '/' + self.name + '/'
		# create backend
		if backend_type == 'caching':
			self.backend = backend.CachingBackend(path='Screenlets/' + p)
		elif backend_type == 'gconf':
			self.backend = backend.GconfBackend()	
		else:
			self.backend = backend.ScreenletsBackend()
			log.error(_("Unable to init backend - settings will not be saved!"))
		
		# get the ids of all saved instances
		self.saved_ids = self.backend.get_configs()		
	
	def create_instance (self, id=None, **keyword_args):
		"""Create a new instance.
		This function returns either the new Screenlet-instance or None.
		Note: the id parameter is unused."""
		log.debug(_("Creating new instance: "))
		if self.saved_ids:
			id = self.saved_ids.pop(0)
			restore = True
		else:
			id = self.__get_next_id()
			restore = False
			
		sl = self.screenlet(id=id, session=self, enable_saving=False, **keyword_args)
		if restore:
			# disable the canvas-updates in the screenlet
			disable = sl.disable_updates
			sl.disable_updates = True
			opts = self.backend.load_instance(sl.id)
			if opts:
				for o in opts:
					opt = sl.get_option_by_name(o)
					if opt:
						setattr(sl, opt.name, opt.on_import(opts[o]))
			# re-enable updates and call redraw/reshape
			sl.disable_updates = disable
			
		sl.enable_saving(True)
		sl.finish_loading()
		self.instances.append(sl) # add screenlet to session
		# and cause initial save to store INI-file in session dir
		sl.width = sl.width
		return sl
	
	def delete_instance (self, id):
		"""Delete the given instance with ID 'id' and remove its session file.
		When the last instance within the session is removed, the session dir 
		is completely removed."""
		sl = self.get_instance_by_id(id)
		if sl:
			# remove instance from session
			if len(self.instances) == 1:
				gtk.main_quit()
			else:
				self.instances.remove(sl)
			
			# delete the instance's settings
			try:
				self.backend.delete_instance(id)
			except Exception:
				log.debug(_("Failed to remove INI-file for instance (not critical)."))
			
			# Delete the config path if all instances have been deleted
			# TODO: This only deletes ~/.config/Screenlets/APPLET_NAME/default
			# We should also delete ~/.config/Screenlets/APPLET_NAME
			if len(self.backend.get_configs()) == 0:
				try:
					os.rmdir(self.backend.path)
				except:
					log.debug(_("Failed to remove session dir '%s' - not empty?") % self.name)			
			return True
		return False
	
	def get_instance_by_id (self, id):
		"""Return the instance with the given id from within this session."""
		for inst in self.instances:
			if inst.id == id:
				return inst
		return None
	
	def quit (self):
		"""quit the session"""
		gtk.main_quit()

	def quit_instance (self, id):
		"""quit the given instance with ID 'id'"""
		sl = self.get_instance_by_id(id)
		if sl:
			# remove instance from session
			if len(self.instances) == 1:
				gtk.main_quit()
			else:
				self.instances.remove(sl)
			self.backend.flush()
			return True
		return False

	def start (self):
		"""Start a new session (or restore an existing session) for the
		current Screenlet-class. Creates a new instance when none is found."""
		# we have a new session running. Indicate that to the system.
		self.__register_screenlet()
		
		log.debug(_("Loading instances in: %s") % self.backend.path)
		# create the first instance
		sl = self.create_instance()
		self.__run_session(sl)
	
	def __register_screenlet (self):
		"""Create new entry for this session in the global TMP_FILE."""
		# if tempfile doesn't exists, create it
		if not self.__create_tempdir():
			return False

		# if screenlet not already added
		running = utils.list_running_applets()
		if running.count(self.screenlet.__name__) == 0:
		# open temp file for appending data
			try:
				f = open(screenlets.TEMP_FILE, 'a') # No need to create a empty file , append will do just fine
			except IOError, e:
				log.error(_("Unable to open %s") % screenlets.TEMP_FILE)
				return False
			else:
				log.debug(_("Creating new entry for %s in %s") % (self.screenlet.__name__, screenlets.TEMP_FILE))
				f.write(self.screenlet.__name__ + '\n')
				f.close()
		else: log.debug(_("Screenlet has already been added to %s") % screenlets.TEMP_FILE)
	
	def __create_tempdir (self):
		"""Create the global temporary file for saving screenlets. The file is 
		used for indicating which screnlets are currently running."""

		# check for existence of screenlets.TEMP_DIR and create it if missing
		if not os.path.isdir(screenlets.TEMP_DIR):
			try:
				if os.path.exists(screenlets.TEMP_DIR):
					# something exists, but is not a directory
					os.remove(screenlets.TEMP_DIR)

				log.debug(_("No global tempdir found, creating new one."))
				os.mkdir(screenlets.TEMP_DIR)
				# make the tmp directory accessible for all users
				
				os.chmod(screenlets.TEMP_DIR, S_IRWXU | S_IRWXG | S_IRWXO)
				log.debug(_('Temp directory %s created.') % screenlets.TEMP_DIR)
			except OSError, e:
				log.error(_('Error: Unable to create temp directory %s - universal-applets-manager will not work properly. \n%s') % (screenlets.TEMP_DIR, e))
				return False
		return True
	
	def __unregister_screenlet (self, name=None):
		"""Delete this session's entry from the gloabl tempfile (and delete the
		entire file if no more running screenlets are set."""
		if not name:
			name = self.screenlet.__name__
		# get running screenlets
		running = utils.list_running_applets()
		if len(running) > 0:
			try:
				running.remove(name)
			except:
				# not found, so ok
				log.debug(_("Entry not found. Will not be removed."))
				return True
			# still running screenlets?
			if running and len(running) > 0:
				# re-save new list of running screenlets
				f = open(screenlets.TEMP_FILE, 'w')
				if f:
					for r in running:
						f.write(r + '\n')
					f.close()
					return True
				else:
					log.error(_("Global tempfile not found."))
				return False
			else:
				log.debug(_('No more screenlets running.'))
				self.__delete_tempfile(name)
		else:
			log.debug(_('No screenlets running?'))
			return False
	
	def __delete_tempfile (self, name=None):
		"""Delete the tempfile for this session."""
		if os.path.isfile(screenlets.TEMP_FILE):
			log.debug(_("Deleting global tempfile %s") % screenlets.TEMP_FILE)
			try:
				os.remove(screenlets.TEMP_FILE)
				return True
			except:
				log.error(_("Error: Failed to delete global tempfile"))
				return False
	
	def __get_next_id (self):
		"""Get the next ID for an instance of the assigned Screenlet.
		TODO: Rewrite this function and avoid creating a new string each time around the loop.
		Python's strings don't perform as well as other datatypes."""
		sln = self.screenlet.__name__[:-9]
		num = 1
		id = sln + str(num)
		while self.get_instance_by_id(id) is not None:
			id = sln + str(num)
			num += 1
		return id
	
	def __run_session (self, main_instance):
		"""Run the session by calling the main handler of the given Screenlet-
		instance. Handles sigkill (?) and keyboard interrupts."""
		# add sigkill-handler
		import signal
		def on_kill(*args):
			pass
		signal.signal(signal.SIGTERM, on_kill)
		
		# set name of tempfile for later (else its missing after kill)
		tempfile = self.screenlet.__name__
		
		# start the main loop
		try:
			gtk.main()
		except KeyboardInterrupt:
			log.info(_("Screenlet '%s' has been interrupted by keyboard.") % self.screenlet.__name__)
			
		# save all unsaved options
		self.backend.flush()
		# remove ourself from the tempfile
		self.__unregister_screenlet(name=tempfile)
	
	def __parse_commandline (self):
		"""Check commandline args for "--session" argument and set session
		name if found. Runs only once during __init__.
		TODO: handle more arguments and maybe allow setting options by
		commandline"""
		import sys
		for arg in sys.argv[1:]:
			# name of session?
			if arg.startswith('--session=') and len(arg)>10:
				self.name = arg[10:]


def create_session (classobj, backend='caching', threading=False):
	"""A very simple utility-function to easily create/start a new session."""
	if threading:
		gtk.gdk.threads_init()
	
	# if this screenlet is already running use DBUS to add a new instance
	sln = classobj.__name__[:-9]
	if services.service_is_running(sln):
		log.info(_("Found a running session of %s, adding new instance by service.") % sln)
		srvc = services.get_service_by_name(sln)
		if srvc:
			log.debug(_("Adding new instance through: %s") % str(srvc))
			srvc.add('')
			return False
			
	# if this is the first instance, create a session
	session = ScreenletSession(classobj, backend_type=backend)
	session.start()
