# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# Screenlets theme module
# (c) 2008 Rico Pfaus (RYX) <ryx@ryxperience.com> and Natan Yellin (Aantn) <aantny@gmail.com>

import glob

import pygtk
pygtk.require('2.0')
import gtk
import gobject
import cairo
import pango
import gettext
try:
	import rsvg
except ImportError:
	print "The rsvg module is not installed. Please install it for improved graphics."

import screenlets
from screenlets.options import *
from screenlets.drawing import Drawing

# translation stuff
gettext.textdomain('screenlets')
gettext.bindtextdomain('screenlets', '/usr/share/locale')

log = screenlets.logger.get_default_logger()

_ = gettext.gettext

class Theme (Drawing):
	"""Themes are a simple way of loading svg and png files from a special theme-directory.
	Every Universal Applet has a theme-directory named "themes". Every individual theme is
	stored in subdirectory of "themes". For example, the theme "yellow" would be stored in
	the directory YOUR_APPLET_DIRECTORY/themes/yellow/
	
	Themes are turned off by default. If your applet uses Themes, just call the
	Screenlet class's constructor with "uses_theme=True".
	
	Although Themes are closely tied into Universal Applets, they can be used in any python
	application. Please see the Theme example in the src/share/examples directory."""
	
	def __init__ (self, path):
		"""Constructor. Raises an exception if the path is invalid.
		path: The path to the theme's directory."""
		
		# meta-info (set through theme.conf)
		self.__name__		= ''	# names
		self.__author__		= ''	# author
		self.__version__	= ''	# version
		self.__info__		= ''	# description
		
		# public attributes
		self.width		= 10			# width of the largest image in the theme
		self.height		= 10			# height of the largest image in the theme
		
		# private attributes
		self.__path 	= path			# path to the theme's directory
		self.__imgs 	= {}			# dictionary of the images in the theme
		self.__external_imgs 	= {}	# dictionary of external images that have been loaded into the theme
		self.__option_overrides = {}	# dictionary of option overrides
		
		# legacy support
		self.path = self.__path
		
		# load all images in the theme's path
		self.loaded = self.__load_all()
		if self.loaded == False:
			log.warning("No files found in theme?")
	
	def __getitem__ (self, item):
		return self.get_entry(item)
	
	def check_entry (self, filename):
		"""Checks if a file with the given name is loaded in this theme.
		filename: The name without an extension.
		return: True if the file is contained in the current theme in either svg or png form."""
		if self.__imgs.has_key(filename+".svg") or self.__imgs.has_key(filename+".png") \
		or self.__external_imgs.has_key(filename+".svg") or self.__external_imgs.has_key(filename+".png"):
			return True
		else:
			return False
	
	def get_entry (self, filename, type="any", internal_only=False):
		"""Get an image that is currently stored in the theme. By default, external images are preferred.
		filename: The name of the file without the extension.
		type: "svg", "png", or "any".
		internal_only: True if only internal images should be used.
		return: The requested image or None."""
		# legacy support
		if filename.endswith(".svg"):
			filename = filename[:-4]
			type = "svg"
		elif filename.endswith(".png"):
			filename = filename[:-4]
			type = "png"
			
		# external images
		if internal_only == False:
			# svgs
			if (type=="svg" or type=="any") and self.__external_imgs.has_key(filename+".svg"):
				return self.__external_imgs[filename+".svg"]
			# pngs
			elif (type=="png" or type=="any") and self.__external_imgs.has_key(filename+".png"):
				return self.__external_imgs[filename+".png"]
		
		# internal images
		# svgs
		if (type=="svg" or type=="any") and self.__imgs.has_key(filename+".svg"):
			return self.__imgs[filename+".svg"]
		# pngs
		elif (type=="png" or type=="any") and self.__imgs.has_key(filename+".png"):
			return self.__imgs[filename+".png"]
		else:
			return None	

	def has_overrides (self):
		"""Check if this theme contains overrides for options."""
		return len(self.__option_overrides) > 0
	
	def get_option_overrides (self):
		"""Get the overrides for this theme."""
		return self.__option_overrides
	
	def load_conf (self, filename):
		"""Load a config-file from this theme's dir and save vars in list."""
		ini = utils.IniReader()
		if ini.load_conf(filename):
			if ini.has_section('Theme'):
				self.__name__ = ini.get_option('name', section='Theme')
				self.__author__ = ini.get_option('author', section='Theme')
				self.__version__ = ini.get_option('version', section='Theme')
				self.__info__ = ini.get_option('info', section='Theme')
			if ini.has_section('Options'):
				opts = ini.list_options(section='Options')
				if opts:
					for o in opts:
						self.__option_overrides[o[0]] = o[1]
			print _("theme.conf loaded: ")
			print _("Name: ") + str(self.__name__)
			print _("Author: ") +str(self.__author__)
			print _("Version: ") +str(self.__version__)
			print _("Info: ") +str(self.__info__)
			return True
		else:
			print _("Failed to load theme.conf")
			return False
	
	def load_svg (self, filename):
		"""Load an SVG-file into this theme."""
		# try to load the svg with the rsvg module
		try:
			file = rsvg.Handle(filename)
			# find the file's size
			if file is not None:
				size = file.get_dimension_data()
				
		# if the rsvg module isn't installed then use gdk
		except NameError, ex:
			file = gtk.gdk.pixbuf_new_from_file(filename)
			# find the file's size
			if file is not None:
				size = (file.get_width, file.get_height)
		
		# set the theme size to the file's size
		if size:
			self.width = size[0]
			self.height = size[1]
		return file
	
	def load_png (self, filename):
		"""Load a PNG-file into this theme."""
		return cairo.ImageSurface.create_from_png(filename)
	
	def __load_all (self):
		"""Load all files in the theme's path. Currently only loads SVGs and
		PNGs.
		Returns True if at least one file was loaded."""
		# clear overrides
		#self.__option_overrides = {}
		
		# create a list of all the files in the theme
		dir_list = glob.glob(self.__path + "/*")
		plen = len(self.__path) + 1
		loaded = False
		
		# loop through the files one by one and load them according to their type
		for file in dir_list:
			img = None
			if file.endswith(".svg"):
				img = self.load_svg(file)
			elif file.endswith(".png"):
				img = self.load_png(file)
			elif file.endswith("theme.conf"):
				print _("theme.conf found! Loading option-overrides.")
				if self.load_conf(file):
					loaded = True
			if img is not None:
				name = file[plen:]
				# TODO: Are the rsvgs automatically freed? (Don't free them with the following code. It causes a segfault.)
				#if name.endswith(".svg") and self.__imgs.has_key(name):
				#	self.__imgs[name].close()
				self.__imgs[name] = img
				loaded = True
		return loaded
	
	def reload (self):
		"""Re-Load all files in the theme's path."""
		self.free()
		self.__load_all()
	
	def free (self):
		"""Deletes the Theme's contents and frees all rsvg-handles."""
		self.__option_overrides.clear()
		#for filename, file in self.__imgs.items():
		#	if filename.endswith(".svg"):
		#			file.close()
		self.__imgs.clear()
	
	# TEST: render-function
	# should be used like "theme.render(context, 'notes-bg')" and then use
	# either an svg or png image
	def render (self, ctx, name):
		"""Render an image from within this theme to the given context. This
		function can EITHER use png OR svg images, so it is possible to 
		create themes using both image-formats when a Screenlet uses this
		function for drawing its images. The image name has to be defined
		without the extension and the function will automatically select 
		the available one (SVG is prefered over PNG)."""
		# render an svg
		if self.__imgs.has_key(name+".svg"):
			# try to render it as a rsvg handle
			try:
				self.__imgs[name+".svg"].render_cairo(ctx)
			# try to render it as a gdk pixbuf if the rsvg module isn't installed
			except:
				ctx.set_source_pixbuf(self.imgs[name+".svg"], 0, 0)	
				ctx.paint()
		# render a png		
		else:
			ctx.set_source_surface(self.__imgs[name+".png"], 0, 0)
			ctx.paint()
	
	def render_png_colorized(self, ctx, name, color): 	
		# Scale the pixmap 	
		ctx.set_source_rgba(color[0], color[1], color[2], color[3]) 	
		ctx.set_source_surface(self.pngs[name+".png"], 0, 0) 	
		ctx.mask_surface(image, 0, 0) 	
		ctx.stroke()
	
	#-----------------------	
	# Deprecated Methods
	#-----------------------
	
	def apply_option_overrides (self, object):
		"""@DEPRECATED Use ThemedApplication.apply_option_overrides()"""
		object.apply_option_overrides()
	
	def show_notification (self,text):
	        """@DEPRECATED Show notification window at current mouse position."""
		if self.notify is None:
	      		self.notify = Notify()
	        	self.notify.text = text
	        	self.notify.show()

	def hide_notification (self):
	        """@DEPRECATED hide notification window"""
		if self.notify is not None:
			self.notify.hide()
			self.notify = None

	def show_tooltip (self,text,tooltipx,tooltipy):
	        """@DEPRECATED Show tooltip window at current mouse position."""
		if self.tooltip is None:
      			self.tooltip = Tooltip(300, 400)
        		self.tooltip.text = text
        		self.tooltip.x    = tooltipx
        		self.tooltip.y    = tooltipy
			self.tooltip.show()

	def hide_tooltip (self):
	        """@DEPRECATED hide tooltip window"""
		if self.tooltip is not None:
			self.tooltip.hide()
			self.tooltip = None
